# shrek llama

import torch
from math import sqrt

from entropix.torch_sampler import sample, device
from .textgen import TextGen, TextChat

class ShrekMixin:
    def sample(self, tokens, **kwargs):
        head_dim = self.model.params['head_dim_kv']
        n_layers = self.model.params['llama.block_count']
        score_name = f'attn{n_layers-1}_pre_scores'

        tokens = torch.tensor(tokens, dtype=torch.int32, device=device)
        logits = self.logits(tokens)
        scores = self.model.get_named_node(score_name) / sqrt(head_dim)

        batch_tokens = tokens.unsqueeze(0)
        batch_logits = logits.unsqueeze(0)
        batch_scores = scores.unsqueeze(0)

        nexts = sample(batch_tokens, batch_logits, batch_scores, **kwargs)
        return nexts.squeeze(0).item()

class ShrekGen(ShrekMixin, TextGen):
    pass

class ShrekChat(ShrekMixin, TextChat):
    pass



# test model

import gadget

print('CPU:linear:numpy')
match = gadget.model.test_linear()

print('CPU:linear:torch')
match = gadget.model.test_linear(framework='torch')

print('CUDA:linear:torch')
match = gadget.model.test_linear(framework='torch', backend='cuda')

print('CPU:getrows:numpy')
match = gadget.model.test_getrows()

print('CPU:getrows:torch')
match = gadget.model.test_getrows(framework='torch')

print('CUDA:getrows:torch')
match = gadget.model.test_getrows(framework='torch', backend='cuda')



from .libs import _libggml

from .libs.constants import(
    GGMLQuantizationType,
    LlamaPoolingType,
)

from .libs._libggml import (
    # initialization
    GGML_DEFAULT_GRAPH_SIZE,
    ggml_init_params,
    ggml_init,
    ggml_free,
    ggml_type_size,
    ggml_tensor_overhead,
    ggml_graph_overhead,
    # backend
    ggml_backend_cpu_init,
    ggml_backend_cuda_init,
    ggml_backend_free,
    ggml_backend_alloc_ctx_tensors,
    ggml_backend_get_default_buffer_type,
    ggml_backend_cpu_set_n_threads,
    ggml_backend_graph_compute,
    ggml_backend_buffer_is_host,
    ggml_backend_tensor_set,
    ggml_backend_tensor_get,
    ggml_backend_tensor_set_async,
    ggml_backend_tensor_get_async,
    # allocation
    ggml_gallocr_new,
    ggml_gallocr_free,
    ggml_gallocr_reserve,
    ggml_gallocr_get_buffer_size,
    ggml_gallocr_alloc_graph,
    # graphs
    ggml_new_graph,
    ggml_build_forward_expand,
    ggml_graph_compute_with_ctx,
    # tensors
    ggml_new_tensor_1d,
    ggml_new_tensor_2d,
    ggml_new_tensor_3d,
    ggml_new_tensor_4d,
    ggml_set_name,
    ggml_is_quantized,
    ggml_is_transposed,
    ggml_is_contiguous,
    ggml_element_size,
    ggml_nelements,
    ggml_nbytes,
    ggml_internal_get_type_traits,
    # tensor ops
    ggml_dup,
    ggml_dup_inplace,
    ggml_add,
    ggml_add_inplace,
    ggml_add_cast,
    ggml_add1,
    ggml_add1_inplace,
    ggml_acc,
    ggml_acc_inplace,
    ggml_sub,
    ggml_sub_inplace,
    ggml_mul,
    ggml_mul_inplace,
    ggml_div,
    ggml_div_inplace,
    ggml_sqr,
    ggml_sqr_inplace,
    ggml_sqrt,
    ggml_sqrt_inplace,
    ggml_log,
    ggml_log_inplace,
    ggml_sum,
    ggml_sum_rows,
    ggml_mean,
    ggml_argmax,
    ggml_repeat,
    ggml_repeat_back,
    ggml_concat,
    ggml_abs,
    ggml_abs_inplace,
    ggml_sgn,
    ggml_sgn_inplace,
    ggml_neg,
    ggml_neg_inplace,
    ggml_step,
    ggml_step_inplace,
    ggml_tanh,
    ggml_tanh_inplace,
    ggml_elu,
    ggml_elu_inplace,
    ggml_relu,
    ggml_leaky_relu,
    ggml_relu_inplace,
    ggml_sigmoid,
    ggml_sigmoid_inplace,
    ggml_gelu,
    ggml_gelu_inplace,
    ggml_gelu_quick,
    ggml_gelu_quick_inplace,
    ggml_silu,
    ggml_silu_inplace,
    ggml_silu_back,
    ggml_hardswish,
    ggml_hardsigmoid,
    ggml_norm,
    ggml_norm_inplace,
    ggml_rms_norm,
    ggml_rms_norm_inplace,
    ggml_group_norm,
    ggml_group_norm_inplace,
    ggml_rms_norm_back,
    ggml_mul_mat,
    ggml_mul_mat_set_prec,
    ggml_mul_mat_id,
    ggml_out_prod,
    ggml_scale,
    ggml_set,
    ggml_set_inplace,
    ggml_scale_inplace,
    ggml_set_1d,
    ggml_set_1d_inplace,
    ggml_set_2d,
    ggml_set_2d_inplace,
    ggml_cpy,
    ggml_cast,
    ggml_cont,
    ggml_cont_1d,
    ggml_cont_2d,
    ggml_cont_3d,
    ggml_cont_4d,
    ggml_reshape,
    ggml_reshape_1d,
    ggml_reshape_2d,
    ggml_reshape_3d,
    ggml_reshape_4d,
    ggml_view_1d,
    ggml_view_2d,
    ggml_view_3d,
    ggml_view_4d,
    ggml_permute,
    ggml_transpose,
    ggml_get_rows,
    ggml_get_rows_back,
    ggml_diag,
    ggml_diag_mask_inf,
    ggml_diag_mask_inf_inplace,
    ggml_diag_mask_zero,
    ggml_diag_mask_zero_inplace,
    ggml_soft_max,
    ggml_soft_max_inplace,
    ggml_soft_max_ext,
    ggml_soft_max_back,
    ggml_soft_max_back_inplace,
    ggml_rope,
    ggml_rope_inplace,
    ggml_rope_ext,
    ggml_rope_ext_inplace,
    ggml_rope_yarn_corr_dims,
    ggml_rope_back,
    ggml_clamp,
    ggml_im2col,
    ggml_conv_depthwise_2d,
    ggml_conv_1d,
    ggml_conv_1d_ph,
    ggml_conv_transpose_1d,
    ggml_conv_2d,
    ggml_conv_2d_sk_p0,
    ggml_conv_2d_s1_ph,
    ggml_conv_transpose_2d_p0,
    ggml_pool_1d,
    ggml_pool_2d,
    ggml_upscale,
    ggml_upscale_ext,
    ggml_pad,
    ggml_timestep_embedding,
    ggml_argsort,
    ggml_arange,
    ggml_top_k,
    ggml_flash_attn_ext,
    ggml_flash_attn_ext_set_prec,
    ggml_flash_attn_back,
    ggml_ssm_conv,
    ggml_ssm_scan,
    ggml_win_part,
    ggml_win_unpart,
    ggml_unary,
    ggml_unary_inplace,
    ggml_get_rel_pos,
    ggml_add_rel_pos,
    ggml_add_rel_pos_inplace,
)



# test llama model

import numpy as np
from gadget.textgen import test_logits, test_logits_hf

# configure
gg_path = '/home/doug/fast/models/meta-llama-3.2-1b-f32.gguf'
hf_model = 'meta-llama/Llama-3.2-1B'

if __name__ == '__main__':
    gg_output = test_logits(gg_path, hf_model, batch_size=32, context_length=1024)
    print(gg_output)
    hf_output = test_logits_hf(hf_model)
    print(hf_output)



# gguf + ggml models

import ast
import numpy as np
from typing import get_type_hints
from collections import defaultdict

from .ggml import GGMLQuantizationType
from .utils import AttrDict
from .loader import GgufFile
from .compute import GgmlCompute

##
## type decorators
##

class Parameter:
    def __init__(self, field):
        self.field = field

class State:
    def __init__(self, field):
        self.field = field

class Tensor:
    def __init__(self, ttype, shape):
        if type(ttype) is str:
            ttype = GGMLQuantizationType[ttype]
        self.ttype = ttype
        self.shape = shape

def resolve_field(key, *dicts):
    if type(key) is str:
        for d in dicts:
            if key in d:
                return d[key]
        raise KeyError(f'key {key} not found')
    else:
        return key

def eval_parameter(expr, gguf):
    if type(expr) is str:
        return gguf.get_field(expr)
    elif callable(expr):
        return expr(gguf)
    return expr

##
## model interface
##

def freeze(func):
    def wrapper(*args, **kwargs):
        return func()
    return wrapper

class GgmlModel(GgmlCompute):
    def __init__(self, params, tensors, states, backend=None, framework=None):
        super().__init__(params, tensors, backend=backend, framework=framework)
        self.state = AttrDict(states)
        self.last_state = None

    @classmethod
    def from_values(cls, values=None, backend=None, framework=None, **params):
        # default empty tensors
        if values is None:
            values = {}

        # get type hints for model
        hints = get_type_hints(cls)

        # get default parameters
        params0 = {k: v.field for k, v in hints.items() if type(v) is Parameter}

        # get state fields
        states = {k: v.field for k, v in hints.items() if type(v) is State}

        # resolve tensor shapes
        tensors = {
            k: (t.ttype, [resolve_field(x, params, params0) for x in t.shape])
            for k, t in hints.items() if type(t) is Tensor
        }

        # create model and graph
        self = cls(params0 | params, tensors, states, backend=backend, framework=framework)

        # set input values
        for k, v in values.items():
            self.set_input(k, v)

        # return model
        return self

    @classmethod
    def from_gguf(cls, gguf, backend=None, framework=None, **params):
        # get metadata from gguf
        weights = {
            key: (ttype, shape)
            for key, (ttype, shape, array) in gguf.tensors.items()
        }

        # get type hints for model
        hints = get_type_hints(cls)

        # get default parameters
        params0 = {
            k: eval_parameter(v.field, gguf)
            for k, v in hints.items() if type(v) is Parameter
        }

        # get state fields
        states = {
            k: eval_parameter(v.field, gguf)
            for k, v in hints.items() if type(v) is State
        }

        # resolve tensor shapes
        tensors = {
            k: (t.ttype, [resolve_field(x, params, params0, gguf.fields) for x in t.shape])
            for k, t in hints.items() if type(t) is Tensor
        }

        # create model and graph
        self = cls(
            gguf.fields | params0 | params, weights | tensors,
            states, backend=backend, framework=framework
        )

        # assign tensors on backend
        for name, (ttype, shape, tensor) in gguf.tensors.items():
            self.set_input(name, tensor)

        # return model
        return self

    @classmethod
    def from_path(cls, path, *args, **kwargs):
        gguf = GgufFile.from_path(path)
        return cls.from_gguf(gguf, *args, **kwargs)

    def forward(self):
        raise NotImplementedError('forward method must be implemented')

    def rebuild_graph(self):
        super().create_graph(freeze(self.forward))

    def __call__(self, **values):
        # check if states have changed
        new_state = tuple(self.state.values())
        if new_state != self.last_state:
            self.last_state = new_state
            self.rebuild_graph()

        # call compute
        return super().__call__(**values)

##
## testing
##

def test_linear(input_dim=64, output_dim=32, batch_size=16, **kwargs):
    from .ggml import ggml_mul_mat, ggml_add

    # simple model interface
    class TestModel(GgmlModel):
        # strings dimensions are filled in dynamically
        x: Tensor('F32', ('input_dim', 'batch_size'))

        def forward(self):
            # get contexts and inputs
            ctx = self.ctx_graph
            a, b, x = self.tensors['a', 'b', 'x']

            # apply function
            x1 = ggml_mul_mat(ctx, a, x, name=f'x1')
            x2 = ggml_add(ctx, x1, b, name=f'x2')

            # return result
            return x2

    # generate weights
    a_np = np.random.randn(output_dim, input_dim).astype(np.float32)
    b_np = np.random.randn(output_dim).astype(np.float32)

    # create dummy gguf
    gguf = GgufFile()
    gguf.set_field('name', b'test')
    gguf.set_field('input_dim', input_dim, dtype=np.int64)
    gguf.set_field('output_dim', output_dim, dtype=np.int64)
    gguf.set_tensor('a', a_np)
    gguf.set_tensor('b', b_np)

    # load gguf as model
    model = TestModel.from_gguf(gguf, batch_size=batch_size, **kwargs)

    # compute on input data
    x_np = np.random.randn(batch_size, input_dim).astype(np.float32)
    y_np = model(x=x_np)

    # bring to host numpy if needed
    if hasattr(y_np, 'numpy'):
        y_np = y_np.cpu().numpy()

    # get numpy results
    y0_np = (x_np @ a_np.T) + b_np[None,:]
    match = np.allclose(y_np, y0_np, atol=1e-5)
    print(match)

    # return result
    return model

def test_getrows(output_dim=32, vocab_size=1024, batch_size=16, **kwargs):
    from .ggml import ggml_get_rows

    # simple model interface
    class TestModel(GgmlModel):
        # strings dimensions are filled in dynamically
        x: Tensor('I32', ('batch_size',))

        def forward(self):
            # get contexts and inputs
            ctx = self.ctx_graph
            m, x = self.tensors['m', 'x']

            # apply function
            x1 = ggml_get_rows(ctx, m, x, name=f'x1')

            # return result
            return x1

    # generate weights
    m_np = np.random.randn(vocab_size, output_dim).astype(np.float32)

    # create dummy gguf
    gguf = GgufFile()
    gguf.set_field('name', b'test')
    gguf.set_field('vocab_size', vocab_size, dtype=np.int64)
    gguf.set_field('output_dim', output_dim, dtype=np.int64)
    gguf.set_tensor('m', m_np)

    # load gguf as model
    model = TestModel.from_gguf(gguf, batch_size=batch_size, **kwargs)

    # compute on input data
    x_np = np.random.randint(0, vocab_size, size=(batch_size,), dtype=np.int32)
    y_np = model(x=x_np)

    # bring to host numpy if needed
    if hasattr(y_np, 'numpy'):
        y_np = y_np.cpu().numpy()

    # get numpy results
    y0_np = m_np.take(x_np, axis=0)
    match = np.allclose(y_np, y0_np, atol=1e-5)
    print(match)

    # return result
    return model



# kv cache class

from ..ggml import ggml_view_3d, ggml_element_size, ggml_cpy, ggml_build_forward_expand
from ..tensor import get_tensor_shape

## kv cache layout (f32)
# k : [n_embd, n_head, n_ctx, n_layer]
# v : [n_embd, n_head, n_ctx, n_layer]

# tensor: [n_embd, n_head, n_ctx, n_layer]
def get_layer_range(ctx, tensor, il, pos0, pos1):
    tsize = ggml_element_size(tensor)
    n_embd, n_head, n_ctx, n_layer = get_tensor_shape(tensor, trim=4)
    length = pos1 - pos0
    offset = (il * n_ctx + pos0) * tsize * n_embd * n_head
    nb1, nb2 = tsize * n_embd, tsize * n_embd * n_head
    return ggml_view_3d(ctx, tensor, n_embd, n_head, length, nb1, nb2, offset)

# src: [n_embd, n_head, length]
# dst: [n_embd, n_head, n_ctx, n_layer]
def set_layer_range(ctx, src, dst, il, pos0, pos1):
    dst1 = get_layer_range(ctx, dst, il, pos0, pos1)
    cpy1 = ggml_cpy(ctx, src, dst1)
    return cpy1

class KVLayerView:
    def __init__(self, cache, ctx, graph, layer, n_past):
        self.cache = cache
        self.ctx = ctx
        self.graph = graph
        self.layer = layer
        self.n_past = n_past

    def get(self, num):
        return self.cache.get_layer_range(self.ctx, self.layer, 0, self.n_past + num)

    def append(self, num, k, v):
        self.cache.set_layer_range(self.ctx, self.graph, self.layer, self.n_past, self.n_past + num, k, v)

    def update(self, k, v):
        _, _, batch_size_k = get_tensor_shape(k, trim=3)
        _, _, batch_size_v = get_tensor_shape(v, trim=3)
        assert batch_size_k == batch_size_v
        batch_size = batch_size_k
        self.append(batch_size, k, v)
        return self.get(batch_size)

class KVCache:
    def __init__(self, k_tensor, v_tensor):
        self.k_tensor = k_tensor
        self.v_tensor = v_tensor

    def get_layer_range(self, ctx, layer, pos0, pos1):
        return (
            get_layer_range(ctx, self.k_tensor, layer, pos0, pos1),
            get_layer_range(ctx, self.v_tensor, layer, pos0, pos1),
        )

    def set_layer_range(self, ctx, graph, layer, pos0, pos1, k, v):
        k1 = set_layer_range(ctx, k, self.k_tensor, layer, pos0, pos1)
        v1 = set_layer_range(ctx, v, self.v_tensor, layer, pos0, pos1)
        ggml_build_forward_expand(graph, k1)
        ggml_build_forward_expand(graph, v1)

    def layer_view(self, ctx, graph, layer, n_past):
        return KVLayerView(self, ctx, graph, layer, n_past)



# utils

import re
from collections import defaultdict, UserDict

# get batch indices
def batch_indices(length, batch_size):
    return [(i, min(i+batch_size, length)) for i in range(0, length, batch_size)]

# split text into chunks
def list_splitter(text, maxlen):
    for i, j in batch_indices(len(text), maxlen):
        yield text[i:j]

def glob_strings(pattern, strings):
    # make regex pattern
    reg = '^' + pattern.replace('.', '\\.').replace('*', '.*') + '$'

    # get matching objects
    matches = [re.match(reg, s) for s in strings]

    # return matching strings
    return [m.string for m in matches if m is not None]

# dictionary that allows getattr access
# only allows string keys for correctness
class AttrDict(UserDict):
    def __setitem__(self, key, val):
        # key type validation
        if type(key) is not str:
            raise ValueError('Only string keys allowed')
        elif '*' in key:
            raise ValueError('Wildcards (*) not allowed in keys')

        # get the key
        super().__setitem__(key, val)

    def __getitem__(self, key):
        # handle tuple accessor case (prevent recursion)
        if type(key) is tuple:
            return [super().__getitem__(k) for k in key]

        # key type validation
        if type(key) is not str:
            raise ValueError('Only string keys allowed')

        # get the key
        return super().__getitem__(key)

    # allows globbing with *
    def subset(self, keys):
        # expand wildcards to lists
        if type(keys) is str:
            keys = glob_strings(keys, self)

        # get subset dict
        return {k: self[k] for k in keys}

# = defaultdict(list)
# + handles popping off maximal list
# + handles deletion on empty list
class SizeDist(dict):
    def __init__(self, data):
        sdist = defaultdict(list)
        for i, size in enumerate(data):
            sdist[size].append(i)
        super().__init__(sdist)

    def pop(self, max_size=None):
        if max_size is None:
            size = max(self, default=None)
        else:
            size = max((s for s in self if s <= max_size), default=None)
        if size is None:
            return
        ret = self[size].pop(0)
        if len(self[size]) == 0:
            del self[size]
        return ret

def pack_batches(sizes, max_len):
    # get size distribution
    n_seq = len(sizes)
    sdist = SizeDist(sizes)
    assert max(sdist) <= max_len

    # plan batch contents
    batches = []
    bidxs = []
    bsize = 0
    for _ in range(n_seq):
        # get a maximal sample
        idx = sdist.pop(max_len-bsize)

        # if none we commit batch and retry
        if idx is None:
            batches.append(bidxs)
            bidxs = []
            bsize = 0
            idx = sdist.pop(max_len)

        # append to batch
        bidxs.append(idx)
        bsize += sizes[idx]

    # append final batch
    batches.append(bidxs)

    return batches



# gadget

from . import ggml
from . import tensor
from . import loader
from . import compute
from . import model
from . import embed
from . import textgen

from .ggml import GGMLQuantizationType, LlamaPoolingType
from .loader import GgufFile
from .compute import GgmlCompute
from .model import GgmlModel, Parameter, State, Tensor
from .embed import EmbedNumpy, EmbedTorch
from .textgen import TextGen, TextChat

from . import models
from .models.bert import BertModel
from .models.llama import LlamaModel

from .ggml import GGMLQuantizationType as T
from .tensor import get_tensor_info

__version__ = "0.4.8"



# global constants and enums

from enum import IntEnum

# file format constants
GGUF_MAGIC             = 0x46554747  # "GGUF"
GGUF_VERSION           = 3
GGUF_DEFAULT_ALIGNMENT = 32
GGML_MAX_DIMS          = 4

# field data types
class GGUFValueType(IntEnum):
    UINT8   = 0
    INT8    = 1
    UINT16  = 2
    INT16   = 3
    UINT32  = 4
    INT32   = 5
    FLOAT32 = 6
    BOOL    = 7
    STRING  = 8
    ARRAY   = 9
    UINT64  = 10
    INT64   = 11
    FLOAT64 = 12

# tensor data types
class GGMLQuantizationType(IntEnum):
    F32     = 0
    F16     = 1
    Q4_0    = 2
    Q4_1    = 3
    Q5_0    = 6
    Q5_1    = 7
    Q8_0    = 8
    Q8_1    = 9
    Q2_K    = 10
    Q3_K    = 11
    Q4_K    = 12
    Q5_K    = 13
    Q6_K    = 14
    Q8_K    = 15
    IQ2_XXS = 16
    IQ2_XS  = 17
    IQ3_XXS = 18
    IQ1_S   = 19
    IQ4_NL  = 20
    IQ3_S   = 21
    IQ2_S   = 22
    IQ4_XS  = 23
    I8      = 24
    I16     = 25
    I32     = 26
    I64     = 27
    F64     = 28
    IQ1_M   = 29
    BF16    = 30

# embed pooling types
class LlamaPoolingType(IntEnum):
    UNSPECIFIED = -1,
    NONE = 0,
    MEAN = 1,
    CLS  = 2,
    LAST = 3,



### ggml bindings

import os
import ctypes
from operator import mul
from itertools import accumulate
from math import prod

from .general import load_shared_lib, ctypes_function, DummyFunction
from .constants import GGMLQuantizationType, GGML_MAX_DIMS

##
## library
##

_ggml = load_shared_lib('libggml.so', 'GADGET_GGML_LIB')

##
## function wrappers
##

ttype_size = {
    GGMLQuantizationType.BF16: 2,
    GGMLQuantizationType.F16: 2,
    GGMLQuantizationType.F32: 4,
    GGMLQuantizationType.F64: 8,
    GGMLQuantizationType.I8: 1,
    GGMLQuantizationType.I16: 2,
    GGMLQuantizationType.I32: 4,
    GGMLQuantizationType.I64: 8,
}

def cumprod(arr):
    return list(accumulate(arr, mul))

def get_tensor_type(tensor):
    value = tensor.contents
    return GGMLQuantizationType(value.type)

def get_tensor_nelem(tensor):
    value = tensor.contents
    return tuple(value.ne[:4])

def get_tensor_strides(tensor):
    value = tensor.contents
    return tuple(value.nb[:4])

def get_tensor_block(tensor):
    value = tensor.contents
    return value.blk_size

def get_arg_info(a):
    if type(a) is ggml_tensor_p:
        ty = get_tensor_type(a)
        ne = get_tensor_nelem(a)
        return f'ggml_tensor: {ty.name} × {ne}'
    else:
        return f'{type(a).__name__}: {a}'

def get_input_info(*args):
    return '\n'.join([str(get_arg_info(a)) for a in args])

# NOTE: this only handles non-quantized tensors
def is_contiguous(tensor):
    ne = get_tensor_nelem(tensor)
    nb = get_tensor_strides(tensor)
    tt = get_tensor_type(tensor)
    ts = ttype_size[tt]
    return nb == tuple(cumprod((ts,) + ne)[:-1])

def named_output(func):
    def wrapper(*args, name=None):
        c = func(*args)
        if name is not None:
            ggml_set_name(c, name.encode('utf-8'))
        return c
    return wrapper

def check_inputs(check):
    def outer(func):
        def inner(ctx, *args):
            if not check(*args):
                raise ValueError(
                    f'{func.__name__}: invalid arguments\n{get_input_info(*args)}'
                )
            return func(ctx, *args)
        return inner
    return outer

##
## pre-emptive type checking and naming
## some of these are copies of inlined functions
##

def ggml_can_mul_mat(t0, t1):
    ne0 = get_tensor_nelem(t0)
    ne1 = get_tensor_nelem(t1)
    nb0 = get_tensor_strides(t0)
    return (
        (ne0[0] == ne1[0]     ) and
        (ne1[2]  % ne0[2] == 0) and
        (ne1[3]  % ne0[3] == 0) and
        (nb0[0] <= nb0[1])
    )

def ggml_can_add(t0, t1):
    ne0 = get_tensor_nelem(t0)
    ne1 = get_tensor_nelem(t1)
    return (
        (ne0[0] % ne1[0] == 0) and
        (ne0[1] % ne1[1] == 0) and
        (ne0[2] % ne1[2] == 0) and
        (ne0[3] % ne1[3] == 0)
    )

def ggml_can_get_rows(t0, t1):
    ne0 = get_tensor_nelem(t0)
    ne1 = get_tensor_nelem(t1)
    ttype1 = get_tensor_type(t1)
    return (
        (ne0[2] == ne1[1]) and
        (ne1[3] == 1     ) and
        (ttype1 == GGMLQuantizationType.I32)
    )

def ggml_can_reshape_3d(t, ne0, ne1, ne2):
    ne = get_tensor_nelem(t)
    return prod(ne) == ne0 * ne1 * ne2

def ggml_can_view_3d(t, ne0, ne1, ne2, nb1, nb2, offset):
    ne = get_tensor_nelem(t)
    return prod(ne) == ne0 * ne1 * ne2

def ggml_can_cpy(t0, t1):
    ne0 = get_tensor_nelem(t0)
    ne1 = get_tensor_nelem(t1)
    return prod(ne0) == prod(ne1)

def ggml_can_soft_max_ext(t, mask, head_wgt, alibi):
    return is_contiguous(mask)

##
## constants
##

GGML_MAX_DIMS      = 4
GGML_MAX_PARAMS    = 2048
GGML_MAX_CONTEXTS  = 64
GGML_MAX_SRC       = 10
GGML_MAX_NAME      = 64
GGML_MAX_OP_PARAMS = 64

GGML_DEFAULT_GRAPH_SIZE = 2048
MAX_FREE_BLOCKS = 256

##
## enums
##

GGML_STATUS_ALLOC_FAILED = -2
GGML_STATUS_FAILED       = -1
GGML_STATUS_SUCCESS      = 0
GGML_STATUS_ABORTED      = 1

# ggml_object_type
GGML_OBJECT_TYPE_TENSOR      = 0
GGML_OBJECT_TYPE_GRAPH       = 1
GGML_OBJECT_TYPE_WORK_BUFFER = 2

##
## structs
##

class ggml_object(ctypes.Structure): ...
ggml_object_p = ctypes.POINTER(ggml_object)
ggml_object._fields_ = [
    ("offs"   , ctypes.c_size_t  ),
    ("size"   , ctypes.c_size_t  ),
    ("next"   , ggml_object_p    ),
    ("type"   , ctypes.c_int     ),
    ("padding", ctypes.c_char * 4),
]

class ggml_scratch(ctypes.Structure):
    _fields_ = [
        ("offs", ctypes.c_size_t),
        ("size", ctypes.c_size_t),
        ("data", ctypes.c_void_p),
    ]

class ggml_init_params(ctypes.Structure):
    _fields_ = [
        ("mem_size"  , ctypes.c_size_t),
        ("mem_buffer", ctypes.c_void_p),
        ("no_alloc"  , ctypes.c_bool  ),
    ]

class ggml_context(ctypes.Structure):
    _fields_ = [
        ("mem_size"        , ctypes.c_size_t),
        ("mem_buffer"      , ctypes.c_void_p),
        ("mem_buffer_owned", ctypes.c_bool  ),
        ("no_alloc"        , ctypes.c_bool  ),
        ("no_alloc_save"   , ctypes.c_bool  ),
        ("n_objects"       , ctypes.c_int   ),
        ("objects_begin"   , ggml_object_p  ),
        ("objects_end"     , ggml_object_p  ),
        ("scratch"         , ggml_scratch   ),
        ("scratch_save"    , ggml_scratch   ),
    ]
ggml_context_p = ctypes.POINTER(ggml_context)

# interlinked tensor/buffer classes
class ggml_tensor                     (ctypes.Structure): ...
class ggml_backend_buffer_i           (ctypes.Structure): ...
class ggml_backend_buffer             (ctypes.Structure): ...
class ggml_backend_buffer_type_i      (ctypes.Structure): ...
class ggml_backend_buffer_type        (ctypes.Structure): ...
class ggml_backend_buffer_type_context(ctypes.Structure): ...

# pointer convenience vars
ggml_tensor_p                      = ctypes.POINTER(ggml_tensor)
ggml_backend_buffer_p              = ctypes.POINTER(ggml_backend_buffer)
ggml_backend_buffer_type_p         = ctypes.POINTER(ggml_backend_buffer_type)
ggml_backend_buffer_context_p      = ctypes.c_void_p
ggml_backend_buffer_type_context_p = ctypes.c_void_p

ggml_backend_buffer_i._fields_ = [
    ("get_name"   , ctypes.CFUNCTYPE(ctypes.c_char_p, ggml_backend_buffer_p                                                                  )),
    ("free_buffer", ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p                                                                  )),
    ("get_base"   , ctypes.CFUNCTYPE(ctypes.c_void_p, ggml_backend_buffer_p                                                                  )),
    ("init_tensor", ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p, ggml_tensor_p                                                   )),
    ("set_tensor" , ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p, ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)),
    ("get_tensor" , ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p, ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)),
    ("cpy_tensor" , ctypes.CFUNCTYPE(ctypes.c_bool  , ggml_backend_buffer_p, ggml_tensor_p, ggml_tensor_p                                    )),
    ("clear"      , ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p, ctypes.c_uint8                                                  )),
    ("reset"      , ctypes.CFUNCTYPE(None           , ggml_backend_buffer_p                                                                  )),
]

ggml_backend_buffer._fields_ = [
    ("iface"  , ggml_backend_buffer_i        ),
    ("buft"   , ggml_backend_buffer_type_p   ),
    ("context", ggml_backend_buffer_context_p),
    ("size"   , ctypes.c_size_t              ),
    ("usage"  , ctypes.c_int                 ),
]

ggml_backend_buffer_type_i._fields_ = [
    ("get_name"      , ctypes.CFUNCTYPE(ctypes.c_char_p      , ggml_backend_buffer_type_p                 )),
    ("alloc_buffer"  , ctypes.CFUNCTYPE(ggml_backend_buffer_p, ggml_backend_buffer_type_p, ctypes.c_size_t)),
    ("get_alignment" , ctypes.CFUNCTYPE(ctypes.c_size_t      , ggml_backend_buffer_type_p                 )),
    ("get_max_size"  , ctypes.CFUNCTYPE(ctypes.c_size_t      , ggml_backend_buffer_type_p                 )),
    ("get_alloc_size", ctypes.CFUNCTYPE(ctypes.c_size_t      , ggml_backend_buffer_type_p, ggml_tensor_p  )),
    ("is_host"       , ctypes.CFUNCTYPE(ctypes.c_bool        , ggml_backend_buffer_type_p                 )),
]

ggml_backend_buffer_type._fields_ = [
    ("iface"  , ggml_backend_buffer_type_i        ),
    ("context", ggml_backend_buffer_type_context_p),
]

GGML_MAX_OP_PARAMS_INT = GGML_MAX_OP_PARAMS // ctypes.sizeof(ctypes.c_int32)
ggml_tensor._fields_ = [
    ("type"     , ctypes.c_int                           ),
    ("buffer"   , ggml_backend_buffer_p                  ),
    ("ne"       , ctypes.c_int64 * GGML_MAX_DIMS         ),
    ("nb"       , ctypes.c_size_t * GGML_MAX_DIMS        ),
    ("op"       , ctypes.c_int                           ),
    ("op_params", ctypes.c_int32 * GGML_MAX_OP_PARAMS_INT),
    ("flags"    , ctypes.c_int32                         ),
    ("grad"     , ggml_tensor_p                          ),
    ("src"      , ggml_tensor_p * GGML_MAX_SRC           ),
    ("view_src" , ggml_tensor_p                          ),
    ("view_offs", ctypes.c_size_t                        ),
    ("data"     , ctypes.c_void_p                        ),
    ("name"     , ctypes.c_char * GGML_MAX_NAME          ),
    ("extra"    , ctypes.c_void_p                        ),
]

# types and quantization
ggml_to_float_p          = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int64)
ggml_from_float_p        = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_float), ctypes.c_void_p, ctypes.c_int64)
ggml_from_float_to_mat_p = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_float), ctypes.c_void_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64)
ggml_vec_dot_p           = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.c_size_t, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int)
ggml_vec_dot_p           = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.c_size_t, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int)
ggml_gemv_p              = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.c_size_t, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
ggml_gemm_p              = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.c_size_t, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int)

class ggml_type_traits(ctypes.Structure):
    _fields_ = [
        ("type_name"           , ctypes.c_char_p         ),
        ("blck_size"           , ctypes.c_int            ),
        ("blck_size_interleave", ctypes.c_int64          ),
        ("type_size"           , ctypes.c_size_t         ),
        ("is_quantized"        , ctypes.c_bool           ),
        ("to_float"            , ggml_to_float_p         ),
        ("from_float"          , ggml_from_float_p       ),
        ("from_float_ref"      , ggml_from_float_p       ),
        ("from_float_to_mat"   , ggml_from_float_to_mat_p),
        ("vec_dot"             , ggml_vec_dot_p          ),
        ("vec_dot_type"        , ctypes.c_int            ),
        ("nrows"               , ctypes.c_int64          ),
        ("ncols"               , ctypes.c_int64          ),
        ("gemv"                , ggml_gemv_p             ),
        ("gemm"                , ggml_gemm_p             ),
    ]
ggml_type_traits_p = ctypes.POINTER(ggml_type_traits)

# graph construction
class ggml_hash_set(ctypes.Structure):
    _fields_ = [
        ("size", ctypes.c_size_t              ),
        ("keys", ctypes.POINTER(ggml_tensor_p)),
    ]

class ggml_cgraph(ctypes.Structure):
    _fields_ = [
        ("size"              , ctypes.c_int                 ),
        ("n_nodes"           , ctypes.c_int                 ),
        ("n_leafs"           , ctypes.c_int                 ),
        ("nodes"             , ctypes.POINTER(ggml_tensor_p)),
        ("grads"             , ctypes.POINTER(ggml_tensor_p)),
        ("leafs"             , ctypes.POINTER(ggml_tensor_p)),
        ("visited_hash_table", ggml_hash_set                ),
        ("order"             , ctypes.c_int                 ),
    ]
ggml_cgraph_p = ctypes.POINTER(ggml_cgraph)

# backend and backend context
class ggml_backend_i    (ctypes.Structure): ...
class ggml_backend      (ctypes.Structure): ...
class ggml_backend_event(ctypes.Structure): ...

# pointer convenience variables
ggml_guid                 = ctypes.c_uint8 * 16
ggml_guid_p               = ctypes.POINTER(ggml_guid)
ggml_backend_p            = ctypes.POINTER(ggml_backend)
ggml_backend_event_p      = ctypes.POINTER(ggml_backend_event)
ggml_backend_context_p    = ctypes.c_void_p
ggml_backend_graph_plan_p = ctypes.c_void_p

ggml_backend_i._fields_ = [
    ("get_name"               , ctypes.CFUNCTYPE(ctypes.c_char_p           , ggml_backend_p                                                                  )),
    ("free"                   , ctypes.CFUNCTYPE(None                      , ggml_backend_p                                                                  )),
    ("get_default_buffer_type", ctypes.CFUNCTYPE(ggml_backend_buffer_type_p, ggml_backend_p                                                                  )),
    ("set_tensor_async"       , ctypes.CFUNCTYPE(None                      , ggml_backend_p, ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)),
    ("get_tensor_async"       , ctypes.CFUNCTYPE(None                      , ggml_backend_p, ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)),
    ("cpy_tensor_async"       , ctypes.CFUNCTYPE(ctypes.c_bool             , ggml_backend_p, ggml_backend_p, ggml_tensor_p, ggml_tensor_p                    )),
    ("synchronize"            , ctypes.CFUNCTYPE(None                      , ggml_backend_p                                                                  )),
    ("graph_plan_create"      , ctypes.CFUNCTYPE(ggml_backend_graph_plan_p , ggml_backend_p, ggml_cgraph_p                                                   )),
    ("graph_plan_free"        , ctypes.CFUNCTYPE(None                      , ggml_backend_p, ggml_backend_graph_plan_p                                       )),
    ("graph_plan_update"      , ctypes.CFUNCTYPE(None                      , ggml_backend_p, ggml_backend_graph_plan_p, ggml_cgraph_p                        )),
    ("graph_plan_compute"     , ctypes.CFUNCTYPE(ctypes.c_int              , ggml_backend_p, ggml_backend_graph_plan_p                                       )),
    ("graph_compute"          , ctypes.CFUNCTYPE(ctypes.c_int              , ggml_backend_p, ggml_cgraph_p                                                   )),
    ("supports_op"            , ctypes.CFUNCTYPE(ctypes.c_bool             , ggml_backend_p, ggml_tensor_p                                                   )),
    ("supports_buft"          , ctypes.CFUNCTYPE(ctypes.c_bool             , ggml_backend_p, ggml_backend_buffer_type_p                                      )),
    ("offload_op"             , ctypes.CFUNCTYPE(ctypes.c_bool             , ggml_backend_p, ggml_tensor_p                                                   )),
    ("event_new"              , ctypes.CFUNCTYPE(ggml_backend_event_p      , ggml_backend_p                                                                  )),
    ("event_free"             , ctypes.CFUNCTYPE(None                      , ggml_backend_event_p                                                            )),
    ("event_record"           , ctypes.CFUNCTYPE(None                      , ggml_backend_event_p                                                            )),
    ("event_wait"             , ctypes.CFUNCTYPE(None                      , ggml_backend_p, ggml_backend_event_p                                            )),
    ("event_synchronize"      , ctypes.CFUNCTYPE(None                      , ggml_backend_event_p                                                            )),
]

ggml_backend._fields_ = [
    ("guid"   , ggml_guid_p           ),
    ("iface"  , ggml_backend_i        ),
    ("context", ggml_backend_context_p),
]

ggml_backend_event._fields_ = [
    ("backend", ggml_backend_p ),
    ("context", ctypes.c_void_p),
]

class hash_node(ctypes.Structure):
    _fields_ = [
        ("n_children", ctypes.c_int   ),
        ("n_views"   , ctypes.c_int   ),
        ("buffer_id" , ctypes.c_int   ),
        ("offset"    , ctypes.c_size_t),
        ("allocated" , ctypes.c_bool  ),
    ]
hash_node_p        = ctypes.POINTER(hash_node)

class tensor_alloc(ctypes.Structure):
    _fields_ = [
        ("buffer_id", ctypes.c_int   ),
        ("offset"   , ctypes.c_size_t),
        ("size_max" , ctypes.c_size_t),
    ]
tensor_alloc_p     = ctypes.POINTER(tensor_alloc)

class leaf_alloc(ctypes.Structure):
    _fields_ = [
        ("buffer_id", ctypes.c_int),
        ("leaf"     , tensor_alloc),
    ]
leaf_alloc_p       = ctypes.POINTER(leaf_alloc)

class node_alloc(ctypes.Structure):
    _fields_ = [
        ("dst", tensor_alloc               ),
        ("src", tensor_alloc * GGML_MAX_SRC),
    ]
node_alloc_p       = ctypes.POINTER(node_alloc)

class free_block(ctypes.Structure):
    _fields_ = [
        ("offset", ctypes.c_size_t),
        ("size"  , ctypes.c_size_t),
    ]
free_block_p       = ctypes.POINTER(free_block)

class ggml_dyn_tallocr(ctypes.Structure):
    _fields_ = [
        ("alignment"    , ctypes.c_size_t             ),
        ("n_free_blocks", ctypes.c_int                ),
        ("free_blocks"  , free_block * MAX_FREE_BLOCKS),
        ("max_size"     , ctypes.c_size_t             ),
    ]
ggml_dyn_tallocr_p = ctypes.POINTER(ggml_dyn_tallocr)

class ggml_gallocr(ctypes.Structure):
    _fields_ = [
        ("bufts"      , ggml_backend_buffer_type_p        ),
        ("buffers"    , ggml_backend_buffer_p             ),
        ("buf_tallocs", ctypes.POINTER(ggml_dyn_tallocr_p)),
        ("n_buffers"  , ctypes.c_int                      ),
        ("hash_set"   , ggml_hash_set                     ),
        ("hash_values", hash_node_p                       ),
        ("node_allocs", node_alloc_p                      ),
        ("n_nodes"    , ctypes.c_int                      ),
        ("leaf_allocs", leaf_alloc_p                      ),
        ("n_leafs"    , ctypes.c_int                      ),
    ]
ggml_gallocr_p = ctypes.POINTER(ggml_gallocr)

##
## functions
##

## initialization

@ctypes_function(_ggml,
    [ggml_init_params],
    ggml_context_p
)
def  ggml_init(params): ...

@ctypes_function(_ggml,
    [ggml_context_p],
    None
)
def ggml_free(ctx): ...

@ctypes_function(_ggml,
    [ctypes.c_int],
    ctypes.c_size_t
)
def ggml_type_size(type): ...

@ctypes_function(_ggml,
    None,
    ctypes.c_size_t,
)
def ggml_tensor_overhead(): ...

@ctypes_function(_ggml,
    None,
    ctypes.c_size_t
)
def ggml_graph_overhead(): ...

## backend

@ctypes_function(_ggml,
    None,
    ggml_backend_p
)
def ggml_backend_cpu_init(): ...

try:
    @ctypes_function(_ggml,
        [ctypes.c_int],
        ggml_backend_p
    )
    def ggml_backend_cuda_init(): ...
except:
    ggml_backend_cuda_init = DummyFunction('CUDA backend not found')

@ctypes_function(_ggml,
    None,
    ggml_backend_p
)
def ggml_backend_free(backend): ...

@ctypes_function(_ggml,
    [ggml_backend_p],
    ggml_backend_buffer_type_p
)
def ggml_backend_get_default_buffer_type(backend): ...

@ctypes_function(_ggml,
    [ggml_backend_p, ctypes.c_int],
    None
)
def ggml_backend_cpu_set_n_threads(backend_cpu, n_threads): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_backend_p],
    ggml_backend_buffer_p
)
def ggml_backend_alloc_ctx_tensors(): ...

@ctypes_function(_ggml,
    [ggml_backend_p, ggml_cgraph_p],
    ctypes.c_int
)
def ggml_backend_graph_compute(backend, cgraph): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t],
    None
)
def ggml_backend_tensor_set(tensor, data, offset, size): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t],
    None
)
def ggml_backend_tensor_get(tensor, data, offset, size): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t],
    None
)
def ggml_backend_tensor_set_async(tensor, data, offset, size): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t],
    None
)
def ggml_backend_tensor_get_async(tensor, data, offset, size): ...

@ctypes_function(_ggml,
    [ggml_backend_buffer_p],
    ctypes.c_bool
)
def ggml_backend_buffer_is_host(buffer): ...

## allocation

@ctypes_function(_ggml,
    [ggml_backend_buffer_type_p],
    ggml_gallocr_p
)
def ggml_gallocr_new(buft): ...

@ctypes_function(_ggml,
    [ggml_gallocr_p],
    None
)
def ggml_gallocr_free(galloc): ...

@ctypes_function(_ggml,
    [ggml_gallocr_p, ggml_cgraph_p],
    ctypes.c_bool
)
def ggml_gallocr_reserve(galloc, graph): ...

@ctypes_function(_ggml,
    [ggml_gallocr_p, ctypes.c_int],
    ctypes.c_size_t
)
def ggml_gallocr_get_buffer_size(galloc, buffer_id): ...

@ctypes_function(_ggml,
    [ggml_gallocr_p, ggml_cgraph_p],
    ctypes.c_bool
)
def ggml_gallocr_alloc_graph(galloc, graph): ...

## tensors

@ctypes_function(_ggml,
    [ggml_context_p, ctypes.c_int, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_new_tensor_1d(ctx, type, ne0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ctypes.c_int, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_new_tensor_2d(ctx, type, ne0, ne1): ...

@ctypes_function(_ggml,
    [ggml_context_p, ctypes.c_int, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_new_tensor_3d(ctx, type, ne0, ne1, ne2): ...

@ctypes_function(_ggml,
    [ggml_context_p, ctypes.c_int, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_new_tensor_4d(ctx, type, ne0, ne1, ne2, ne3): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_char_p],
    None
)
def ggml_set_name(tensor, name): ...

@ctypes_function(_ggml,
    [ggml_tensor_p],
    ctypes.c_size_t
)
def ggml_element_size(tensor): ...

@ctypes_function(_ggml,
    [ggml_tensor_p],
    ctypes.c_int64
)
def ggml_nelements(tensor): ...

@ctypes_function(_ggml,
    [ggml_tensor_p],
    ctypes.c_int64
)
def ggml_nbytes(tensor): ...

@ctypes_function(_ggml,
    [ctypes.c_int],
    ctypes.c_bool
)
def ggml_is_quantized(tensor): ...

@ctypes_function(_ggml,
    [ggml_tensor_p],
    ctypes.c_bool
)
def ggml_is_transposed(tensor): ...

@ctypes_function(_ggml,
    [ggml_tensor_p],
    ctypes.c_bool
)
def ggml_is_contiguous(tensor): ...

@ctypes_function(_ggml,
    [ctypes.c_int],
    ggml_type_traits
)
def ggml_internal_get_type_traits(ttype): ...

## graphs

@ctypes_function(_ggml,
    [ggml_context_p],
    ggml_cgraph_p
)
def ggml_new_graph(ctx): ...

@ctypes_function(_ggml,
    [ggml_cgraph_p, ggml_tensor_p],
)
def ggml_build_forward_expand(cgraph, tensor): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_cgraph_p, ctypes.c_int],
    ctypes.c_int
)
def ggml_graph_compute_with_ctx(ctx, cgraph, n_threads): ...

## tensor ops

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_dup(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_dup_inplace(ctx, a): ...

@named_output
@check_inputs(ggml_can_add)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_add_cast(ctx, a, b, type): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add1(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add1_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_acc(ctx, a, b, nb1, nb2, nb3, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_acc_inplace(ctx, a, b, nb1, nb2, nb3, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sub(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sub_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_mul(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_mul_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_div(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_div_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sqr(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sqr_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sqrt(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sqrt_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_log(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_log_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sum(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sum_rows(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_mean(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_argmax(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_repeat(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_repeat_back(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_concat(ctx, a, b, dim): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_abs(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_abs_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sgn(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sgn_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_neg(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_neg_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_step(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_step_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_tanh(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_tanh_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_elu(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_elu_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_relu(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float, ctypes.c_bool],
    ggml_tensor_p
)
def ggml_leaky_relu(ctx, a, negative_slope, inplace): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_relu_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sigmoid(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_sigmoid_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_gelu(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_gelu_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_gelu_quick(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_gelu_quick_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_silu(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_silu_inplace(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_silu_back(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_hardswish(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_hardsigmoid(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_norm(ctx, a, eps): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_norm_inplace(ctx, a, eps): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rms_norm(ctx, a, eps): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rms_norm_inplace(ctx, a, eps): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_group_norm(ctx, a, n_groups): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_group_norm_inplace(ctx, a, n_groups): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rms_norm_back(ctx, a, b, eps): ...

@named_output
@check_inputs(ggml_can_mul_mat)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_mul_mat(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_int],
    None
)
def ggml_mul_mat_set_prec(a, prec): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_mul_mat_id(ctx, as_, b, ids): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_out_prod(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_scale(ctx, a, s): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float],
    ggml_tensor_p
)
def ggml_scale_inplace(ctx, a, s): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set(ctx, a, b, nb1, nb2, nb3, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set_inplace(ctx, a, b, nb1, nb2, nb3, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set_1d(ctx, a, b, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set_1d_inplace(ctx, a, b, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set_2d(ctx, a, b, nb1, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_set_2d_inplace(ctx, a, b, nb1, offset): ...

@named_output
@check_inputs(ggml_can_cpy)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_cpy(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_cast(ctx, a, type): ...

@named_output
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_cont(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_cont_1d(ctx, a, ne0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_cont_2d(ctx, a, ne0, ne1): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_cont_3d(ctx, a, ne0, ne1, ne2): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_cont_4d(ctx, a, ne0, ne1, ne2, ne3): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_reshape(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_reshape_1d(ctx, a, ne0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_reshape_2d(ctx, a, ne0, ne1): ...

@named_output
@check_inputs(ggml_can_reshape_3d)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_reshape_3d(ctx, a, ne0, ne1, ne2): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64],
    ggml_tensor_p
)
def ggml_reshape_4d(ctx, a, ne0, ne1, ne2, ne3): ...

@named_output
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_view_1d(ctx, a, ne0, offset): ...

@named_output
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_view_2d(ctx, a, ne0, ne1, nb1, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_view_3d(ctx, a, ne0, ne1, ne2, nb1, nb2, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_int64, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t],
    ggml_tensor_p
)
def ggml_view_4d(ctx, a, ne0, ne1, ne2, ne3, nb1, nb2, nb3, offset): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_permute(ctx, a, axis0, axis1, axis2, axis3): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_transpose(ctx, a): ...

@named_output
@check_inputs(ggml_can_get_rows)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_get_rows(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_get_rows_back(ctx, a, b, c): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_diag(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_diag_mask_inf(ctx, a, n_past): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_diag_mask_inf_inplace(ctx, a, n_past): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_diag_mask_zero(ctx, a, n_past): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_diag_mask_zero_inplace(ctx, a, n_past): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_soft_max(ctx, a): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_soft_max_inplace(ctx, a): ...

@named_output
@check_inputs(ggml_can_soft_max_ext)
@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_soft_max_ext(ctx, a, mask, scale, max_bias): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_soft_max_back(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_soft_max_back_inplace(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_rope(ctx, a, b, n_dims, mode): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_rope_inplace(ctx, a, b, n_dims, mode): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rope_ext(ctx, a, b, c, n_dims, mode, n_ctx_orig, freq_base, freq_scale, ext_factor, attn_factor, beta_fast, beta_slow): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rope_ext_inplace(ctx, a, b, c, n_dims, mode, n_ctx_orig, freq_base, freq_scale, ext_factor, attn_factor, beta_fast, beta_slow): ...

@ctypes_function(_ggml,
    [ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.POINTER(ctypes.c_float)],
    None
)
def ggml_rope_yarn_corr_dims(n_dims, n_ctx_orig, freq_base, beta_fast, beta_slow, dims): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_rope_back(ctx, a, b, c, n_dims, mode, n_ctx_orig, freq_base, freq_scale, ext_factor, attn_factor, beta_fast, beta_slow): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_clamp(ctx, a, min, max): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_bool, ctypes.c_int],
    ggml_tensor_p
)
def ggml_im2col(ctx, a, b, s0, s1, p0, p1, d0, d1, is_2D, dst_type): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_depthwise_2d(ctx, a, b, s0, s1, p0, p1, d0, d1): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_1d(ctx, a, b, s0, p0, d0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_1d_ph(ctx, a, b, s, d): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_transpose_1d(ctx, a, b, s0, p0, d0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_2d(ctx, a, b, s0, s1, p0, p1, d0, d1): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_conv_2d_sk_p0(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_conv_2d_s1_ph(ctx, a, b): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_conv_transpose_2d_p0(ctx, a, b, stride): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_pool_1d(ctx, a, op, k0, s0, p0): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_pool_2d(ctx, a, op, k0, k1, s0, s1, p0, p1): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_upscale(ctx, a, scale_factor): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_upscale_ext(ctx, a, ne0, ne1, ne2, ne3): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_pad(ctx, a, p0, p1, p2, p3): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_timestep_embedding(ctx, timesteps, dim, max_period): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_argsort(ctx, a, order): ...

@ctypes_function(_ggml,
    [ggml_context_p, ctypes.c_float, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_arange(ctx, start, stop, step): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_top_k(ctx, a, k): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_float, ctypes.c_float],
    ggml_tensor_p
)
def ggml_flash_attn_ext(ctx, q, k, v, mask, scale, max_bias): ...

@ctypes_function(_ggml,
    [ggml_tensor_p, ctypes.c_int],
    None
)
def ggml_flash_attn_ext_set_prec(a, prec): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ctypes.c_bool],
    ggml_tensor_p
)
def ggml_flash_attn_back(ctx, q, k, v, d, masked): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_ssm_conv(ctx, s, x, c, sq): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_ssm_scan(ctx, s, x, dt, A, B, C, sq): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_win_part(ctx, a, w): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_win_unpart(ctx, a, w0, h0, w): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_unary(ctx, a, op): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int],
    ggml_tensor_p
)
def ggml_unary_inplace(ctx, a, op): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ctypes.c_int, ctypes.c_int],
    ggml_tensor_p
)
def ggml_get_rel_pos(ctx, a, qh, kh): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add_rel_pos(ctx, a, pw, ph): ...

@ctypes_function(_ggml,
    [ggml_context_p, ggml_tensor_p, ggml_tensor_p, ggml_tensor_p],
    ggml_tensor_p
)
def ggml_add_rel_pos_inplace(ctx, a, pw, ph): ...



# test llama textgen

from gadget.textgen import test_textgen

# configure
gg_path = '/home/doug/fast/models/meta-llama-3.2-1b-f32.gguf'
hf_model = 'meta-llama/Llama-3.2-1B'
prompt = 'The capital of France is'

if __name__ == '__main__':
    test_textgen(gg_path, hf_model, prompt=prompt, context_length=1024, max_gen=512, backend='cuda')



# llama implementation

import numpy as np

from ..ggml import (
    ggml_element_size,
    ggml_add_inplace,
    ggml_get_rows,
    ggml_transpose,
    ggml_view_1d,
    ggml_view_2d,
    ggml_cont,
)
from ..tensor import get_tensor_shape, get_tensor_info
from ..model import GgmlModel, Parameter, State, Tensor
from .cache import KVCache
from .layers import (
    linear_layer,
    norm_layer,
    attention_layer,
    feed_forward_layer,
)

##
## llama model
##

def get_head_dim_kv(gguf):
    n_head_kv = gguf.get_field('llama.attention.head_count_kv')
    embed_size_kv = gguf.get_tensor_shape('blk.0.attn_k.weight')[1]
    assert embed_size_kv % n_head_kv == 0
    return embed_size_kv // n_head_kv

def causal_mask(context_length, batch_size):
    minctx = int(batch_size) - int(context_length) # cast to int to avoid overflow
    ctxpos = np.arange(minctx, batch_size, dtype=np.int32)
    posids = np.arange(batch_size, dtype=np.int32)
    mask   = np.where(ctxpos[None, :] <= posids[:, None], 0.0, -np.inf).astype(np.float32)
    return mask

def clip_mask(ctx, mask, n_past, n_tokens, name=None):
    context_length, batch_size = get_tensor_shape(mask)
    mask_tsize = ggml_element_size(mask)
    mask_stride = mask_tsize * context_length
    mask_offset = mask_tsize * (context_length - n_past - batch_size)
    mask = ggml_view_2d(ctx, mask, n_past + n_tokens, n_tokens, mask_stride, mask_offset)
    return ggml_cont(ctx, mask, name=name)

class LlamaModel(GgmlModel):
    batch_size    : Parameter(512)
    context_length: Parameter('llama.context_length')
    head_dim_kv   : Parameter(get_head_dim_kv)

    n_past  : State(0)
    n_tokens: State(None)

    tokens   : Tensor('I32', ('context_length',))
    positions: Tensor('I32', ('context_length',))
    mask     : Tensor('F32', ('context_length', 'batch_size'))

    kcache   : Tensor('F32', ('head_dim_kv', 'llama.attention.head_count_kv', 'context_length', 'llama.block_count'))
    vcache   : Tensor('F32', ('head_dim_kv', 'llama.attention.head_count_kv', 'context_length', 'llama.block_count'))

    # perform param validation here
    def __init__(self, params, tensors, states, **kwargs):
        # validate batch_size and context_length
        if (bs := params['batch_size']) > (cl := params['context_length']):
            raise ValueError('batch_size ({bs}) > context_length ({cl})')
        if (cl := params['context_length']) > (cl0 := params['llama.context_length']):
            raise ValueError('context_length ({cl}) > maximum context_length ({cl0})')

        # pass to model constructor
        super().__init__(params, tensors, states, **kwargs)

        # set position and mask tensors
        self.set_input('positions', np.arange(cl, dtype=np.int32))
        self.set_input('mask', causal_mask(cl, bs))

        # make kv cache
        self.kv_cache = KVCache(self.tensors['kcache'], self.tensors['vcache'])

    def __call__(self, tokens):
        # accept a raw list
        if type(tokens) is list:
            tokens = np.array(tokens, dtype=np.int32)

        # set token batch size
        self.state['n_tokens'] = len(tokens)
        n_past, n_tokens = self.state['n_past', 'n_tokens']

        # set inputs and evaluate model
        self.set_input('tokens', tokens, offset=n_past)
        output = super().__call__()

        # update state
        self.state['n_past'] += n_tokens

        # return output
        return output

    def reset(self):
        # reset kv_cache position
        self.state['n_past'] = 0

    # llama model function
    def forward(self):
        ctx = self.ctx_graph

        # get runtime state
        batch_size, context_length = self.params['batch_size', 'context_length']
        n_past, n_tokens = self.state['n_past', 'n_tokens']

        # get params
        n_layers, n_heads_q, n_heads_kv, rope_base, layer_norm_rms_eps = self.params[
            'llama.block_count'            , 'llama.attention.head_count',
            'llama.attention.head_count_kv', 'llama.rope.freq_base'      ,
            'llama.attention.layer_norm_rms_epsilon',
        ]

        # select used input tokens
        tokens, positions, mask = self.tensors['tokens', 'positions', 'mask']
        tsize, psize = ggml_element_size(tokens), ggml_element_size(positions)
        tokens = ggml_view_1d(ctx, tokens, n_tokens, tsize * n_past, name='tokens_batch')
        positions = ggml_view_1d(ctx, positions, n_tokens, psize * n_past, name='positions_batch')
        mask = clip_mask(ctx, mask, n_past, n_tokens, name='mask_batch')

        # get token embeddings
        etok, rope_freqs = self.tensors['token_embd.weight', 'rope_freqs.weight']
        cur = ggml_get_rows(ctx, etok, tokens, name='embed=tok')

        # loop over layers
        for i in range(n_layers):
            # get layer tensors
            wq, wk, wv, wo, wan, wu, wd, wg, wn, = self.tensors[
                f'blk.{i}.attn_q.weight'     , f'blk.{i}.attn_k.weight'   , f'blk.{i}.attn_v.weight'  ,
                f'blk.{i}.attn_output.weight', f'blk.{i}.attn_norm.weight', f'blk.{i}.ffn_up.weight'  ,
                f'blk.{i}.ffn_down.weight'   , f'blk.{i}.ffn_gate.weight' , f'blk.{i}.ffn_norm.weight',
            ]

            # get attention interactions
            cache = self.kv_cache.layer_view(ctx, self.graph, i, n_past)
            att = norm_layer(ctx, cur, wan, rms=True, eps=layer_norm_rms_eps, name=f'attn{i}_norm')
            att = attention_layer(
                ctx, att, n_heads_q, mask, wq, wk, wv, wo, positions=positions, n_heads_kv=n_heads_kv,
                rope_freqs=rope_freqs, rope_base=rope_base, eps=layer_norm_rms_eps, kv_cache=cache,
                name=f'attn{i}'
            )

            # add layer input to attention
            att = ggml_add_inplace(ctx, att, cur)

            # feed forward network on current
            cur = norm_layer(ctx, att, wn, rms=True, eps=layer_norm_rms_eps, name=f'ffn{i}_norm')
            cur = feed_forward_layer(ctx, cur, wg, wd, wg=wu, act='silu', name=f'ffn{i}') # notice wg/wu flipped

            # add attention output to current tensor
            cur = ggml_add_inplace(ctx, cur, att)

        # get output tensors
        onw = self.tensors['output_norm.weight']
        ow = self.tensors.get('output.weight', etok)

        # generate output
        cur = norm_layer(ctx, cur, onw, rms=True, eps=layer_norm_rms_eps, name='output_norm')
        cur = linear_layer(ctx, cur, ow, name='output')

        # return logits
        return cur



# test memory usage

import tracemalloc
import gadget

tracemalloc.start()

print(tracemalloc.get_traced_memory())

gg = gadget.compute.test_torch()

print(tracemalloc.get_traced_memory())

del gg

print(tracemalloc.get_traced_memory())



# test compute interface

import gadget
from gadget.ggml import GGMLQuantizationType as T

print('CPU:F32:numpy')
match = gadget.compute.test_numpy(qtype=T.F32)

print('CPU:F32:torch')
match = gadget.compute.test_torch(qtype=T.F32)

print('CPU:F16:numpy')
match = gadget.compute.test_numpy(qtype=T.F16)

print('CPU:F16:torch')
match = gadget.compute.test_torch(qtype=T.F16)

print('CPU:Q8_0:numpy')
match = gadget.compute.test_numpy(qtype=T.Q8_0)

print('CPU:Q8_0:torch')
match = gadget.compute.test_torch(qtype=T.Q8_0)

print('CPU:Q4_0:numpy')
match = gadget.compute.test_numpy(qtype=T.Q4_0)

print('CPU:Q4_0:torch')
match = gadget.compute.test_torch(qtype=T.Q4_0)

print('CUDA:F32:torch')
match = gadget.compute.test_torch(qtype=T.F32, backend='cuda')

print('CUDA:F16:torch')
match = gadget.compute.test_torch(qtype=T.F16, backend='cuda')



# simple ggml test

import gc
import ctypes
import numpy as np

from .ggml import (
    ggml_type_size,
    ggml_tensor_overhead,
    ggml_graph_overhead,
    ggml_init_params,
    ggml_init,
    ggml_new_tensor_1d,
    ggml_new_tensor_2d,
    ggml_mul_mat,
    ggml_add,
    ggml_new_graph,
    ggml_build_forward_expand,
    ggml_graph_compute_with_ctx,
    ggml_backend_cpu_init,
    ggml_backend_alloc_ctx_tensors,
    ggml_free,
    ggml_backend_get_default_buffer_type,
    ggml_gallocr_new,
    ggml_gallocr_reserve,
    ggml_gallocr_get_buffer_size,
    ggml_gallocr_alloc_graph,
    ggml_backend_cpu_set_n_threads,
    ggml_backend_graph_compute,
    GGMLQuantizationType,
    GGML_DEFAULT_GRAPH_SIZE,
)

##
## type conversion
##

gtype_to_ctype = {
    GGMLQuantizationType.F32: ctypes.c_float,
    # GGMLQuantizationType.F16: ctypes.c_half, # not supported by ctypes
    GGMLQuantizationType.I8: ctypes.c_int8,
    GGMLQuantizationType.I16: ctypes.c_int16,
    GGMLQuantizationType.I32: ctypes.c_int32,
    GGMLQuantizationType.I64: ctypes.c_int64,
}

gtype_to_dtype = {
    GGMLQuantizationType.F32: np.float32,
    # GGMLQuantizationType.F16: np.float16, # not supported by ctypes
    GGMLQuantizationType.I8: np.int8,
    GGMLQuantizationType.I16: np.int16,
    GGMLQuantizationType.I32: np.int32,
    GGMLQuantizationType.I64: np.int64,
}

##
## tensor utilities
##

def trim_nelem(shape):
    dims = 1 + max([
        i for i, d in enumerate(shape) if d > 1
    ], default=0)
    return shape[:dims]

def get_tensor_name(tensor):
    value = tensor.contents
    return value.name.decode('utf-8')

def get_tensor_shape(tensor, raw=False):
    value = tensor.contents
    nelem = tuple(value.ne[:4])
    return trim_nelem(nelem)[::-1]

def get_tensor_type(tensor):
    value = tensor.contents
    return GGMLQuantizationType(value.type)

def get_tensor_info(tensor):
    name = get_tensor_name(tensor)
    ttype = get_tensor_type(tensor)
    shape = get_tensor_shape(tensor)
    stat = f'{name}: {ttype.name} × {shape}'
    return stat

# this assumes the data is contiguous
# will implicity squeeze unit dimensions
def array_to_tensor(array, tensor):
    # check ctype support
    ttype = get_tensor_type(tensor)
    if ttype not in gtype_to_ctype:
        raise ValueError(f'unsupported type: {ttype}')

    # check dtype match
    dtype = gtype_to_dtype[ttype]
    if array.dtype != dtype:
        raise ValueError(f'array dtype mismatch: {array.dtype} != {dtype}')

    # get data pointers
    src = array.ctypes.data
    dst = tensor.contents.data
    size = array.nbytes

    # copy data
    ctypes.memmove(dst, src, size)

# this makes a new array and copies
# we want to avoid deallocating ggml buffers
def tensor_to_array(tensor):
    # check ctype support
    ttype = get_tensor_type(tensor)
    if ttype not in gtype_to_dtype:
        raise ValueError(f'unsupported type: {ttype}')

    # get data pointers
    src = tensor.contents.data
    shape = get_tensor_shape(tensor)
    dtype = gtype_to_dtype[ttype]

    # create numpy array
    array = np.empty(shape, dtype=dtype)
    dst = array.ctypes.data
    size = array.nbytes

    # copy data
    ctypes.memmove(src, dst, size)

    # return array
    return array

def test_context():
    # tensor shapes
    rows_A, cols_A = 4, 2
    rows_B, cols_B = 3, 2

    # compute memory requirements
    mem_usage = (
        ggml_type_size(GGML_TYPE_F32) * (rows_A * cols_A) +
        ggml_type_size(GGML_TYPE_F32) * (rows_B * cols_B) +
        ggml_tensor_overhead() +
        ggml_graph_overhead() +
        1024
    )

    # initialize ggml
    params = ggml_init_params(mem_usage, None, False)
    ctx = ggml_init(params)

    # create tensors
    a = ggml_new_tensor_2d(ctx, GGML_TYPE_F32, cols_A, rows_A)
    b = ggml_new_tensor_2d(ctx, GGML_TYPE_F32, cols_B, rows_B)

    # do multiplication
    c = ggml_mul_mat(ctx, a, b)

    # create graph
    gf = ggml_new_graph(ctx)
    ggml_build_forward_expand(gf, c)

    # create numpy arrays
    a_np = tensor_to_numpy(a)
    b_np = tensor_to_numpy(b)

    # fill arrays
    a_np[:,:] = [[2 , 8], [5, 1], [4, 2], [8, 6]]
    b_np[:,:] = [[10, 5], [9, 9], [5, 4]]

    # compute graph
    ggml_graph_compute_with_ctx(ctx, gf, 1)

    # get results
    n_nodes = gf.contents.n_nodes
    c_tensor = gf.contents.nodes[n_nodes-1]
    c_np = tensor_to_numpy(c_tensor)

    # test results
    assert np.allclose(c_np, (a_np @ b_np.T).T)

    # free memory
    ggml_free(ctx)

    # return arrays
    return a_np, b_np, c_np

def test_backend():
    # tensor shapes
    n_layers, embed_dim, batch_size = 50, 32, 16

    # initialize cpu backend
    backend = ggml_backend_cpu_init()

    # create tensor context
    num_tensors = 1 + 2 * n_layers # input + weights + biases
    mem_tensors = ggml_tensor_overhead() * num_tensors
    par_tensors = ggml_init_params(mem_tensors, None, True)
    ctx_tensors = ggml_init(par_tensors)

    # create tensors
    inputs = ggml_new_tensor_2d(ctx_tensors, GGMLQuantizationType.F32, embed_dim, batch_size)
    weights, biases = zip(*[
        (
            ggml_new_tensor_2d(ctx_tensors, GGMLQuantizationType.F32, embed_dim, embed_dim),
            ggml_new_tensor_1d(ctx_tensors, GGMLQuantizationType.F32, embed_dim)
        ) for _ in range(n_layers)
    ])

    # assign tensors to backend
    ggml_backend_alloc_ctx_tensors(ctx_tensors, backend)

    # compute memory requirements for graph
    mem_graph = (
        ggml_tensor_overhead() * GGML_DEFAULT_GRAPH_SIZE +
        ggml_graph_overhead()
    )
    arr_graph = ctypes.create_string_buffer(mem_graph)
    buf_graph = ctypes.cast(arr_graph, ctypes.c_void_p)

    # create graph context
    par_graph = ggml_init_params(mem_graph, buf_graph, True)
    ctx_graph = ggml_init(par_graph)
    graph = ggml_new_graph(ctx_graph)

    # create network function
    x = inputs
    for i in range(n_layers):
        x = ggml_mul_mat(ctx_graph, weights[i], x, name=f'a{i}')
        x = ggml_add(ctx_graph, x, biases[i], name=f'b{i}')
    outputs = x

    # build graph from network
    ggml_build_forward_expand(graph, outputs)

    # allocate buffers for graph (worst case scenario)
    buf_type = ggml_backend_get_default_buffer_type(backend)
    allocr = ggml_gallocr_new(buf_type)
    ggml_gallocr_reserve(allocr, graph)
    ggml_gallocr_alloc_graph(allocr, graph)
    mem_worst = ggml_gallocr_get_buffer_size(allocr, 0)
    print(f'compute buffer size: {mem_worst/1024:.4f} KiB')

    # set backend runtime options
    ggml_backend_cpu_set_n_threads(backend, 1)

    # set weights and biases
    for i in range(n_layers):
        weight_np = np.random.randn(embed_dim, embed_dim).astype(np.float32)
        biase_np = np.random.randn(embed_dim).astype(np.float32)
        array_to_tensor(weight_np, weights[i])
        array_to_tensor(biase_np, biases[i])

    # set input data
    inputs_np = np.random.randn(batch_size, embed_dim).astype(np.float32)
    array_to_tensor(inputs_np, inputs)

    # do computation
    ggml_backend_graph_compute(backend, graph)

    # get results
    outputs_np = tensor_to_array(outputs)

    # return locals()
    return inputs_np, outputs_np



# test bert model

import numpy as np
from gadget.embed import test_embed

# configure
gg_path = '/home/doug/fast/embed'
hf_model = 'TaylorAI/bge-micro-v2'

# run tests
print('CPU:F32:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-f32.gguf', hf_model)

print('CPU:Q8_0:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-q8_0.gguf', hf_model)

print('CPU:Q4_K_M:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-q4_k_m.gguf', hf_model)

print('CUDA:F32:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-f32.gguf', hf_model, backend='cuda')

print('CUDA:Q8_0:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-q8_0.gguf', hf_model, backend='cuda')

print('CUDA:Q4_K_M:bge-micro')
test_embed(f'{gg_path}/bge-micro-v2-q4_k_m.gguf', hf_model, backend='cuda')



# bert implementation

import numpy as np

from ..ggml import (
    ggml_add,
    ggml_add_inplace,
    ggml_get_rows,
    ggml_view_1d,
    ggml_view_2d,
    ggml_cont,
    ggml_element_size,
)
from ..model import GgmlModel, Parameter, State, Tensor
from .layers import (
    linear_layer,
    norm_layer,
    attention_layer,
    feed_forward_layer,
)

##
## bert model
##

class BertModel(GgmlModel):
    batch_size: Parameter('bert.context_length')
    n_tokens  : State(None)
    tokens    : Tensor('I32', ('batch_size',))
    positions : Tensor('I32', ('batch_size',))
    mask     : Tensor('F32', ('batch_size', 'batch_size'))

    # perform param validation here
    def __init__(self, params, tensors, states, **kwargs):
        # validate batch_size
        if (bs := params['batch_size']) > (cl := params['bert.context_length']):
            raise ValueError('batch_size ({bs}) > context_length ({cl})')

        # pass to model constructor
        super().__init__(params, tensors, states, **kwargs)

    # update state with number of tokens
    # NOTE: this still needs inputs of size batch_size
    #       because we can't set 2d input tensor slices yet
    def __call__(self, tokens, positions, mask, n_tokens=None):
        self.state['n_tokens'] = n_tokens if n_tokens is not None else len(tokens)
        return super().__call__(tokens=tokens, positions=positions, mask=mask)

    # bert model function
    def forward(self):
        ctx = self.ctx_graph

        # get params
        n_layers, n_heads, embed_dim, batch_size, layer_norm_eps = self.params[
            'bert.block_count'     , 'bert.attention.head_count',
            'bert.embedding_length', 'batch_size'               ,
            'bert.attention.layer_norm_epsilon'
        ]

        # get weights
        etok, etyp, epos, tnw, tnb = self.tensors[
            'token_embd.weight', 'token_types.weight', 'position_embd.weight',
            'token_embd_norm.weight', 'token_embd_norm.bias',
        ]

        # get state
        n_tokens = self.state['n_tokens']

        # get input tensors
        tokens, positions, mask = self.tensors['tokens', 'positions', 'mask']

        # get just this batch of tokens
        mask_stride = ggml_element_size(mask) * batch_size
        tokens = ggml_view_1d(ctx, tokens, n_tokens, 0, name='tokens_batch')
        positions = ggml_view_1d(ctx, positions, n_tokens, 0, name='positions_batch')
        mask = ggml_cont(ctx, ggml_view_2d(ctx, mask, n_tokens, n_tokens, mask_stride, 0, name='mask_batch'))

        # get token embeddings (token+type+position+norm)
        cur = ggml_get_rows(ctx, etok, tokens, name='embed=tok')
        cur = ggml_add_inplace(ctx, cur, ggml_view_1d(ctx, etyp, embed_dim, 0), name='embed=tok+typ')
        cur = ggml_add_inplace(ctx, cur, ggml_get_rows(ctx, epos, positions), name='embed=tok+typ+pos')
        cur = norm_layer(ctx, cur, tnw, tnb, eps=layer_norm_eps, inplace=True, name='embed_norm')

        # loop over layers
        for i in range(n_layers):
            # get layer tensors
            wq, bq, wk, bk, wv, bv, wo, bo, wan, ban, wu, bu, wd, bd, wln, bln = self.tensors[
                f'blk.{i}.attn_q.weight'           , f'blk.{i}.attn_q.bias'           ,
                f'blk.{i}.attn_k.weight'           , f'blk.{i}.attn_k.bias'           ,
                f'blk.{i}.attn_v.weight'           , f'blk.{i}.attn_v.bias'           ,
                f'blk.{i}.attn_output.weight'      , f'blk.{i}.attn_output.bias'      ,
                f'blk.{i}.attn_output_norm.weight' , f'blk.{i}.attn_output_norm.bias' ,
                f'blk.{i}.ffn_up.weight'           , f'blk.{i}.ffn_up.bias'           ,
                f'blk.{i}.ffn_down.weight'         , f'blk.{i}.ffn_down.bias'         ,
                f'blk.{i}.layer_output_norm.weight', f'blk.{i}.layer_output_norm.bias',
            ]

            # get attention interactions
            att = attention_layer(
                ctx, cur, n_heads, mask, wq, wk, wv, wo, bq=bq, bk=bk, bv=bv, bo=bo,
                eps=layer_norm_eps, name=f'attn{i}'
            )

            # add attention output to current then normalize
            att = ggml_add_inplace(ctx, cur, att)
            att = norm_layer(ctx, att, wan, ban, eps=layer_norm_eps, inplace=True, name=f'attn{i}_norm')

            # feed forward network on current
            cur = feed_forward_layer(ctx, att, wu, wd, bu=bu, bd=bd, act='gelu', name=f'ffn{i}')

            # add attention output to current tensor and normalize
            cur = ggml_add_inplace(ctx, cur, att, name=f'add{i}')
            cur = norm_layer(ctx, cur, wln, bln, eps=layer_norm_eps, inplace=True, name=f'norm{i}')

        # return embeddings
        return cur



# common layers for reuse

from math import sqrt

from ..ggml import (
    LlamaPoolingType,
    ggml_norm,
    ggml_norm_inplace,
    ggml_rms_norm,
    ggml_rms_norm_inplace,
    ggml_mul,
    ggml_mul_inplace,
    ggml_add,
    ggml_add_inplace,
    ggml_gelu,
    ggml_silu,
    ggml_rope_ext,
    ggml_mul_mat,
    ggml_permute,
    ggml_transpose,
    ggml_cont,
    ggml_cont_2d,
    ggml_reshape_2d,
    ggml_reshape_3d,
    ggml_soft_max_ext,
)
from ..tensor import get_tensor_shape

def linear_layer(ctx, x, weight, bias=None, name=None):
    x = ggml_mul_mat(ctx, weight, x, name=f'{name}_mul')
    if bias is not None:
        x = ggml_add_inplace(ctx, x, bias, name=f'{name}_add')
    return x

def norm_layer(ctx, x, weight, bias=None, eps=0.0, rms=False, inplace=False, name=None):
    if inplace:
        norm_func = ggml_rms_norm_inplace if rms else ggml_norm_inplace
        mul_func, add_func = ggml_mul_inplace, ggml_add_inplace
    else:
        norm_func = ggml_rms_norm if rms else ggml_norm
        mul_func, add_func = ggml_mul, ggml_add
    x = norm_func(ctx, x, eps, name=f'{name}_norm')
    x = mul_func(ctx, x, weight, name=f'{name}_mul')
    if bias is not None:
        x = add_func(ctx, x, bias, name=f'{name}_add')
    return x

def rope_extended(
    ctx, x, pos, n_dims, freqs=None, mode=0, n_ctx_orig=0, freq_base=10000.0, freq_scale=1.0,
    ext_factor=0.0, attn_factor=1.0, beta_fast=0.0, beta_slow=0.0, inplace=True
):
    return ggml_rope_ext(
        ctx, x, pos, freqs, n_dims, mode, n_ctx_orig, freq_base, freq_scale,
        ext_factor, attn_factor, beta_fast, beta_slow, inplace
    )

def attention_layer(
    ctx, x, n_heads, mask, wq, wk, wv, wo, bq=None, bk=None, bv=None, bo=None, n_heads_kv=None,
    rope_freqs=None, rope_base=None, eps=0.0, positions=None, alibi=0.0, kv_cache=None, name=None
):
    # get n_heads_q and n_heads_kv
    n_heads_q = n_heads
    if n_heads_kv is None:
        n_heads_kv = n_heads

    # get dimensions
    embed_dim, batch_size = get_tensor_shape(x, trim=2)
    _, embed_dim_q = get_tensor_shape(wq, trim=2)
    _, embed_dim_k = get_tensor_shape(wk, trim=2)
    _, embed_dim_v = get_tensor_shape(wv, trim=2)

    # kv consistency
    if embed_dim_v != embed_dim_k:
        raise ValueError(f'embed_dim_v ({embed_dim_v}) must be equal to embed_dim_k ({embed_dim_k})')
    embed_dim_kv = embed_dim_k

    # check head divisibility
    if embed_dim_q % n_heads_q != 0:
        raise ValueError(f'embed_dim_q ({embed_dim_q}) must be divisble by n_heads_q ({n_heads_q})')
    if embed_dim_kv % n_heads_kv != 0:
        raise ValueError(f'embed_dim_kv ({embed_dim_kv}) must be divisble by n_heads_kv ({n_heads_kv})')

    # head dims match
    head_dim_q = embed_dim_q // n_heads_q
    head_dim_kv = embed_dim_kv // n_heads_kv
    if head_dim_q != head_dim_kv:
        raise ValueError(f'head_dim_q ({head_dim_q}) must be equal to head_dim_kv ({head_dim_kv})')
    head_dim = head_dim_q

    # compute query, key, value
    q = linear_layer(ctx, x, wq, bias=bq)
    k = linear_layer(ctx, x, wk, bias=bk)
    v = linear_layer(ctx, x, wv, bias=bv)

    # reshape to head_dim
    q = ggml_reshape_3d(ctx, q, head_dim, n_heads_q, batch_size, name=f'{name}_q')
    k = ggml_reshape_3d(ctx, k, head_dim, n_heads_kv, batch_size, name=f'{name}_k')
    v = ggml_reshape_3d(ctx, v, head_dim, n_heads_kv, batch_size, name=f'{name}_v')

    # apply rotary position embeddings
    if rope_base is not None:
        q = rope_extended(ctx, q, pos=positions, n_dims=head_dim, freqs=rope_freqs, freq_base=rope_base)
        k = rope_extended(ctx, k, pos=positions, n_dims=head_dim, freqs=rope_freqs, freq_base=rope_base)

    # apply kv cache
    if kv_cache is not None:
        k, v = kv_cache.update(k, v)

    # permute dimensions
    q = ggml_permute(ctx, q, 0, 2, 1, 3)
    k = ggml_cont(ctx, ggml_permute(ctx, k, 0, 2, 1, 3))

    # compute interactions
    head_wgt = 1.0/sqrt(head_dim)
    kq = ggml_mul_mat(ctx, k, q, name=f'{name}_pre_scores')
    kq = ggml_soft_max_ext(ctx, kq, mask, head_wgt, alibi, name=f'{name}_scores')

    # pull in values
    v = ggml_cont(ctx, ggml_permute(ctx, v, 1, 2, 0, 3))
    kqv = ggml_mul_mat(ctx, v, kq)

    # merge dimensions
    kqv = ggml_permute(ctx, kqv, 0, 2, 1, 3)
    kqv = ggml_cont_2d(ctx, kqv, embed_dim, batch_size)

    # apply output layer
    out = linear_layer(ctx, kqv, wo, bias=bo, name=f'{name}_out')

    # return output
    return out

activations = {
    'gelu': ggml_gelu,
    'silu': ggml_silu,
}

# without gate layer, just a plain feed forward network
# with gate layer, it's a gated feed forward network
def feed_forward_layer(ctx, x, wu, wd, wg=None, bu=None, bd=None, bg=None, act='gelu', name=None):
    y = linear_layer(ctx, x, wu, bias=bu, name=f'{name}_up')
    y = activations[act](ctx, y)
    if wg is not None:
        g = linear_layer(ctx, x, wg, bias=bg, name=f'{name}_gate')
        y = ggml_mul_inplace(ctx, y, g)
    y = linear_layer(ctx, y, wd, bias=bd, name=f'{name}_down')
    return y



# high level embedding interface

import time
import numpy as np
from transformers import AutoTokenizer

from .ggml import LlamaPoolingType
from .models.bert import BertModel

# we don't want to require torch for gadget
try:
    import torch
except ImportError:
    pass

##
## embed
##

class EmbedBase:
    def __init__(self, gguf_path, model_id, pooling=None, model_class=BertModel, **kwargs):
        self.toker = AutoTokenizer.from_pretrained(model_id)
        self.model = model_class.from_path(gguf_path, **kwargs)
        self.batch_size = self.model.params['batch_size']

        # assign pooling type
        if pooling is None:
            pooling = self.model.params.get('bert.pooling_type', LlamaPoolingType.NONE)

        self.pooling = LlamaPoolingType(pooling)

    def tokenize(self, texts):
        if type(texts) is str:
            texts = [texts]
        return self.toker(texts)['input_ids']

    def embed(self, texts, pooling=None, normalize=True):
        # get tokens as list of lists
        tokens = self.tokenize(texts)
        total = sum([len(t) for t in tokens])

        # check for batch fit
        if total > self.batch_size:
            raise ValueError(f'Number of tokens ({total}) > batch_size ({self.batch_size})')

        # create input arrays and compute
        tokids, posids, seqids, mask = self.prepare_inputs(tokens)
        embeds = self.model(tokids, posids, mask, n_tokens=total)

        # do requested pooling
        pooling = self.pooling if pooling is None else pooling
        seqids = seqids.to(embeds.device) if self.model.framework == 'torch' else seqids
        embeds = self.pool_embeds(pooling, embeds[:total, :], seqids[:total])

        # return embeddings
        if normalize:
            embeds = self.norm_embeds(embeds)
        return embeds

class EmbedNumpy(EmbedBase):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, framework='numpy', **kwargs)

    @staticmethod
    def norm_embeds(embeds):
        return embeds / np.linalg.norm(embeds, axis=-1, keepdims=True)

    # this assumes that sequences are packed in order
    @staticmethod
    def pool_embeds(pooling, embeds, seqids):
        if pooling == LlamaPoolingType.NONE:
            _, first = np.unique(seqids, return_index=True)
            pooled = np.split(embeds, first)
        elif pooling == LlamaPoolingType.MEAN:
            uids, ntoks = np.unique(seqids, return_counts=True)
            weights = (uids[:, None] == seqids[None, :]).astype(np.float32)
            weights /= ntoks[:, None]
            pooled = weights @ embeds
        elif pooling == LlamaPoolingType.CLS:
            _, first = np.unique(seqids, return_index=True)
            pooled = embeds[first, :]
        elif pooling == LlamaPoolingType.LAST:
            _, rlast = np.unique(seqids[::-1], return_index=True)
            last = len(seqids) - rlast - 1
            pooled = embeds[last, :]
        else:
            raise ValueError('must specify pooling type')
        return pooled

    def prepare_inputs(self, tokens):
        ntoks = np.array([len(ts) for ts in tokens], dtype=np.int32)
        nseqs, total = len(tokens), ntoks.sum()
        padding = np.zeros(self.batch_size - total, dtype=np.int32)
        tokens = np.concatenate([*(np.array(t, dtype=np.int32) for t in tokens), padding])
        posits = np.concatenate([*(np.arange(n, dtype=np.int32) for n in ntoks), padding])
        seqids = np.concatenate([np.repeat(np.arange(nseqs, dtype=np.int32), ntoks), padding - 1])
        mask = np.where(seqids[:, None] == seqids[None, :], 0.0, -np.inf).astype(np.float32)
        return tokens, posits, seqids, mask

class EmbedTorch(EmbedBase):
    def __init__(self, *args, **kwargs):
        return super().__init__(*args, framework='torch', **kwargs)

    @staticmethod
    def norm_embeds(embeds):
        return embeds / embeds.norm(dim=-1, keepdim=True)

    # this assumes ordered and contiguous seqids
    @staticmethod
    def first_indices(values):
        uvals, indices = values.unique(return_inverse=True)
        first = torch.empty(len(uvals), device=values.device, dtype=torch.int32)
        order = torch.arange(len(values), device=values.device, dtype=torch.int32)
        first.scatter_reduce_(0, indices, order, reduce='amin')
        return first

    # this assumes that sequences are packed in order
    @classmethod
    def pool_embeds(cls, pooling, embeds, seqids):
        if pooling == LlamaPoolingType.NONE:
            first = cls.first_indices(seqids)
            pooled = torch.split(embeds, first, 0)
        elif pooling == LlamaPoolingType.MEAN:
            uids, ntoks = torch.unique(seqids, return_counts=True)
            weights = (uids[:, None] == seqids[None, :]).float()
            weights /= ntoks[:, None]
            pooled = weights @ embeds
        elif pooling == LlamaPoolingType.CLS:
            first = cls.first_indices(seqids)
            pooled = embeds[first, :]
        elif pooling == LlamaPoolingType.LAST:
            rlast = cls.first_indices(seqids[::-1])
            last = len(seqids) - rlast - 1
            pooled = embeds[last, :]
        else:
            raise ValueError('must specify pooling type')
        return pooled

    def prepare_inputs(self, tokens):
        ntoks = torch.tensor([len(ts) for ts in tokens], dtype=torch.int32)
        nseqs, total = len(tokens), ntoks.sum()
        padding = torch.zeros(self.batch_size - total, dtype=torch.int32)
        tokens = torch.cat([*(torch.tensor(t, dtype=torch.int32) for t in tokens), padding])
        posits = torch.cat([*(torch.arange(n, dtype=torch.int32) for n in ntoks), padding])
        seqids = torch.cat([torch.repeat_interleave(torch.arange(nseqs, dtype=torch.int32), ntoks), padding - 1])
        mask = torch.where(seqids[:, None] == seqids[None, :], 0.0, -torch.inf).float()
        return tokens, posits, seqids, mask

##
## test
##

def test_embed(gguf_path, model_id, prompt='hello world', embed_class=EmbedTorch, model_class=BertModel, **kwargs):
    from transformers import AutoModel

    # embed with gg model
    gg_model = embed_class(gguf_path, model_id, model_class=model_class, **kwargs)
    gg_embed = gg_model.embed(prompt)

    # bring to host numpy if needed
    if hasattr(gg_embed, 'numpy'):
        gg_embed = gg_embed.cpu().numpy()

    # embed with hf
    hf_toker = AutoTokenizer.from_pretrained(model_id)
    hf_model = AutoModel.from_pretrained(model_id)
    hf_input = hf_toker(prompt, return_tensors='pt')['input_ids']
    hf_seqid = torch.zeros_like(hf_input)
    with torch.no_grad():
        hf_state = hf_model(hf_input).last_hidden_state[0]
    hf_poold = EmbedTorch.pool_embeds(gg_model.pooling, hf_state, hf_seqid)
    hf_embed = hf_poold / hf_poold.norm(dim=-1, keepdim=True)
    hf_embed = hf_embed.cpu().numpy()

    # embed with ggml
    simil = (hf_embed * gg_embed).sum()
    print(simil)

    return gg_model

def profile_embed(gguf_path, model_id, prompt=None, length=256, reps=25, embed_class=EmbedTorch, model_class=BertModel, **kwargs):
    if prompt is None:
        prompt = ' '.join('a' for _ in range(length-2))

    t0 = time.time()
    total = reps * length

    gg_model = embed_class(gguf_path, model_id, model_class=model_class, **kwargs)

    t1 = time.time()
    d1 = t1 - t0
    print(f'load: {d1}')

    for _ in range(reps):
        gg_embed = gg_model.embed(prompt)

    t2 = time.time()
    d2 = t2 - t1
    print(f'embd: {d2}')
    print(f'rps : {reps/d2}')
    print(f'tps : {total/d2}')



### general stuff

import os
import ctypes

##
## missing utils
##

class DummyFunction:
    def __init__(self, message):
        self.message = message

    def __call__(self, *args, **kwargs):
        raise Exception(self.message)

##
## ctypes helpers
##

# load a shared lib with env override
def load_shared_lib(lib_name, env_var=None):
    # get shared library path
    if env_var is not None and env_var in os.environ:
        lib_path = os.environ[env_var]
    else:
        module_path = os.path.dirname(os.path.abspath(__file__))
        lib_path = os.path.join(module_path, lib_name)

    # load shared library
    try:
        lib_obj = ctypes.CDLL(lib_path)
    except Exception as e:
        raise RuntimeError(f"Failed to load shared library '{lib_name}': {e}")

    # return library object
    return lib_obj

# decorate a function with ctypes information
def ctypes_function(library, argtypes=None, restype=None):
    if argtypes is None:
        argtypes = []
    def decorator(func):
        name = func.__name__
        func = getattr(library, name)
        func.argtypes = argtypes
        func.restype = restype
        return func
    return decorator

##
## libc
##

_libc = ctypes.CDLL(None)

@ctypes_function(_libc,
    [ctypes.c_size_t],
    ctypes.c_void_p
)
def malloc(size): ...

@ctypes_function(_libc,
    [ctypes.c_void_p]
)
def free(ptr): ...



# text generation

import numpy as np
from transformers import AutoTokenizer

from .loader import GgufFile
from .models.llama import LlamaModel

# we don't want to require torch for gadget
try:
    import torch
except ImportError:
    pass

def load_model(gguf_or_path, model_class, **kwargs):
    if type(gguf_or_path) is str:
        return model_class.from_path(gguf_or_path, **kwargs)
    elif type(gguf_or_path) is GgufFile:
        return model_class.from_gguf(gguf_or_path, **kwargs)
    else:
        raise ValueError('must specify gguf file or path')

def sample(logits, temperature=0.7, top_p=0.9, top_k=50):
    probs = torch.exp(logits / temperature)
    probs = probs / torch.sum(probs)
    return torch.multinomial(probs, num_samples=1).item()

def sprint(text):
    print(text, end='', flush=True)

class TextGen:
    def __init__(self, gguf_path, model_id, model_class=LlamaModel, **kwargs):
        self.model = load_model(gguf_path, model_class, framework='torch', **kwargs)
        self.toker = AutoTokenizer.from_pretrained(model_id)

    def reset(self):
        self.model.reset()

    def tokenize(self, texts, **kwargs):
        return self.toker(texts, **kwargs)['input_ids']

    def detokenize(self, tokens, **kwargs):
        return self.toker.decode(tokens, **kwargs)

    def logits(self, tokens):
        return torch.atleast_2d(self.model(tokens))

    def sample(self, tokens, **kwargs):
        logits = self.logits(tokens)
        return sample(logits[-1,:], **kwargs)

    def stream_tokens(self, tokens, max_gen=128, **kwargs):
        batch = tokens
        for _ in range(max_gen):
            tok = self.sample(batch, **kwargs)
            batch = [tok]
            yield tok

    def stream(self, text, **kwargs):
        tokens = self.tokenize(text)
        for tok in self.stream_tokens(tokens, **kwargs):
            yield self.detokenize([tok], add_special_tokens=False)

    def generate(self, text, **kwargs):
        tokens = self.tokenize(text)
        for tok in self.stream_tokens(tokens, **kwargs):
            tokens += [tok]
        return self.detokenize(tokens)

class TextChat(TextGen):
    def __init__(self, gguf_path, model_id, system=None, **kwargs):
        super().__init__(gguf_path, model_id, **kwargs)
        self.system = system
        self.reset(system)

    def reset(self, system=None):
        super().reset()
        system = self.system if system is None else system
        if system is not None:
            self.history = [{'role': 'system', 'content': system}]
        else:
            self.history = []

    def stream_chat(self, message, prefill=None, add_generation_prompt=True, **kwargs):
        self.history.append({'role': 'user', 'content': message})
        n_past = self.model.state['n_past']
        tokens_full = self.toker.apply_chat_template(self.history, add_generation_prompt=add_generation_prompt)
        if prefill is not None:
            tokens_full += self.tokenize(prefill, add_special_tokens=False)
            reply = prefill
        else:
            reply = ''
        tokens = tokens_full[n_past:]
        for tok in self.stream_tokens(tokens, **kwargs):
            if tok == self.toker.eos_token_id:
                break
            text = self.detokenize([tok], add_special_tokens=False)
            reply += text
            yield text
        self.history.append({'role': 'assistant', 'content': reply})

    def generate_chat(self, message, **kwargs):
        tokens = []
        for tok in self.stream_chat(message, **kwargs):
            tokens += [tok]
        return self.detokenize(tokens)

def test_logits(gguf_path, model_id, model_class=LlamaModel, batch_size=128, **kwargs):
    model = TextGen(gguf_path, model_id, model_class=model_class, batch_size=batch_size, **kwargs)
    prompt = 'The capital of France is'
    tokes = model.tokenize(prompt)
    logits = model.logits(tokes)
    return logits

def test_logits_hf(model_id, **kwargs):
    from transformers import AutoModelForCausalLM
    hf_prompt = 'The capital of France is'
    hf_toker = AutoTokenizer.from_pretrained(model_id)
    hf_model = AutoModelForCausalLM.from_pretrained(model_id, output_hidden_states=True)
    hf_input = hf_toker(hf_prompt, return_tensors='pt')['input_ids']
    hf_pos = torch.arange(len(hf_input[0])).unsqueeze(0)
    with torch.no_grad():
        hf_state = hf_model(hf_input)[0]
    return hf_state[0]

def test_textgen(
    gguf_path, model_id, prompt='The capital of France is', model_class=LlamaModel, batch_size=128,
    max_gen=128, **kwargs
):
    model = TextGen(gguf_path, model_id, model_class=model_class, batch_size=batch_size, **kwargs)
    for tok in model.stream(prompt, max_gen=max_gen):
        sprint(tok)



# higher level ggml interface

import re
import ctypes
import numpy as np

from .utils import AttrDict
from .ggml import (
    GGML_DEFAULT_GRAPH_SIZE,
    GGMLQuantizationType as T,
    ggml_tensor_overhead,
    ggml_graph_overhead,
    ggml_init_params,
    ggml_init,
    ggml_free,
    ggml_new_graph,
    ggml_build_forward_expand,
    ggml_backend_cpu_init,
    ggml_backend_cuda_init,
    ggml_backend_free,
    ggml_backend_alloc_ctx_tensors,
    ggml_backend_get_default_buffer_type,
    ggml_backend_cpu_set_n_threads,
    ggml_backend_graph_compute,
    ggml_gallocr_new,
    ggml_gallocr_reserve,
    ggml_gallocr_alloc_graph,
    ggml_gallocr_free,
)
from .tensor import (
    get_framework,
    get_tensor_name,
    get_tensor_info,
    create_tensor,
    array_to_tensor,
    tensor_to_array,
)

##
## compute interface
##

class GgmlCompute:
    def __init__(self, params, tensors, model=None, backend=None, framework=None):
        # initialize empty
        self.backend = None
        self.ctx_tensors = None

        # other options
        self.framework = 'numpy' if framework is None else framework

        # create tensors and graph
        self.create_params(params)
        self.create_backend(backend)
        self.create_tensors(tensors)

        # create graph
        self.alloc = None
        self.ctx_graph = None
        self.arr_graph = None
        self.graph = None
        self.output = None
        if model is not None:
            self.create_graph(model)

    def __del__(self):
        self.destroy_graph()
        if self.ctx_tensors is not None:
            ggml_free(self.ctx_tensors)
            self.ctx_tensors = None
        if self.backend is not None:
            ggml_backend_free(self.backend)
            self.backend = None

    def create_params(self, params):
        self.params = AttrDict(params)

    def create_backend(self, name):
        if name is None or name == 'cpu':
            self.backend = ggml_backend_cpu_init()
            self.backend_type = 'cpu'
        elif (reg := re.match(r'^cuda(?::(\d+))?$', name)) is not None:
            num, = reg.groups()
            num = 0 if num is None else str(num)
            self.backend = ggml_backend_cuda_init(num)
            self.backend_type = 'cuda'
        else:
            raise ValueError(f'unknown backend: {name}')

    def create_tensors(self, specs):
        # create tensor context
        num_tensors = len(specs)
        mem_tensors = ggml_tensor_overhead() * num_tensors
        par_tensors = ggml_init_params(mem_tensors, None, True)
        self.ctx_tensors = ggml_init(par_tensors)

        # create tensors
        self.tensors = AttrDict({
            nam: create_tensor(self.ctx_tensors, typ, shp, nam=nam)
            for nam, (typ, shp) in specs.items()
        })

        # assign tensors on backend
        self.backend_buf = ggml_backend_alloc_ctx_tensors(self.ctx_tensors, self.backend)

    def get_tensor(self, tensor, framework=None, device=None):
        if framework is None:
            framework = self.framework
        if device is None:
            device = self.backend_type
        return tensor_to_array(tensor, framework=framework, device=device)

    # get tensor values as numpy (copy)
    def get_input(self, name, **kwargs):
        tensor = self.tensors[name]
        return self.get_tensor(tensor, **kwargs)

    # set tensor values using numpy
    # offset is in terms of dimension=0 elements
    def set_input(self, name, array, offset=None):
        tensor = self.tensors[name]
        strict = offset is None
        offset = 0 if offset is None else offset
        try:
            array_to_tensor(array, tensor, offset=offset, strict=strict)
        except ValueError as e:
            raise ValueError(f'error setting input "{name}":\n{e}')

    def get_node(self, index, **kwargs):
        n_nodes = self.graph.contents.n_nodes
        if index >= n_nodes:
            raise ValueError(f'index ({index}) >= n_nodes ({n_nodes})')
        tensor = self.graph.contents.nodes[index]
        return self.get_tensor(tensor, **kwargs)

    def get_named_node(self, name, **kwargs):
        n_nodes = self.graph.contents.n_nodes
        for i in range(n_nodes):
            tensor = self.graph.contents.nodes[i]
            tname = get_tensor_name(tensor)
            if tname == name:
                return self.get_tensor(tensor, **kwargs)
        raise ValueError(f'node named "{name}" not found')

    # create computational graph
    def create_graph(self, model, graph_size=GGML_DEFAULT_GRAPH_SIZE):
        # destroy old graph
        self.destroy_graph()

        # compute memory requirements for graph
        # NOTE: need to keep reference to arr_graph around to prevent garbage collect!!!
        mem_graph = (
            ggml_graph_overhead() + ggml_tensor_overhead() * graph_size
        )
        self.arr_graph = ctypes.create_string_buffer(mem_graph)

        # create graph context (this stores tensor/buffer metadata)
        buf_graph = ctypes.cast(self.arr_graph, ctypes.c_void_p)
        par_graph = ggml_init_params(mem_graph, buf_graph, True)
        self.ctx_graph = ggml_init(par_graph)

        # create graph and expand (this creates a graph of ggml operations)
        self.graph = ggml_new_graph(self.ctx_graph)
        self.output = model(self.ctx_graph, self.params, self.tensors)
        ggml_build_forward_expand(self.graph, self.output)

        # allocate buffers for graph (worst case scenario)
        buf_type = ggml_backend_get_default_buffer_type(self.backend)
        self.alloc = ggml_gallocr_new(buf_type)

        # allocate tensors to buffers for graph
        ggml_gallocr_reserve(self.alloc, self.graph)
        ggml_gallocr_alloc_graph(self.alloc, self.graph)

    def destroy_graph(self):
        if self.alloc is not None:
            ggml_gallocr_free(self.alloc)
            self.alloc = None
        if self.ctx_graph is not None:
            ggml_free(self.ctx_graph)
            self.ctx_graph = None
        if self.arr_graph is not None:
            del self.arr_graph
            self.arr_graph = None
        if self.graph is not None:
            self.graph = None
        if self.output is not None:
            self.output = None

    # do computation
    def compute(self):
        ggml_backend_graph_compute(self.backend, self.graph)

    def __call__(self, **values):
        # set input values
        for name, value in values.items():
            self.set_input(name, value)

        # run that baby
        self.compute()

        # get results
        output_np = self.get_tensor(self.output)

        # return results
        return output_np

    def __repr__(self):
        name = self.__class__.__name__
        lines = (
            [f'{name}(backend={self.backend_type})'] + ['', 'INPUTS'] +
            [get_tensor_info(tensor) for tensor in self.tensors.values()]

        )
        if self.graph is not None:
            graph = self.graph.contents
            lines += ['', 'GRAPH'] + [
                get_tensor_info(graph.nodes[i]) for i in range(graph.n_nodes)
            ]

        return '\n'.join(lines)

##
## testing
##

def test_numpy(input_dim=256, output_dim=32, batch_size=16, qtype=T.F32):
    from .ggml import ggml_mul_mat, ggml_add

    # model parameters
    params = dict(
        input_dim=input_dim, output_dim=output_dim, batch_size=batch_size
    )

    # tensor specifications
    tensors = dict(
        a = (qtype, (input_dim, output_dim)),
        b = (T.F32, (output_dim,)),
        x = (T.F32, (input_dim, batch_size)),
    )

    # define model function
    def test_model(ctx, par, ten):
        n, m = par['input_dim'], par['output_dim']
        a, b, x = ten['a'], ten['b'], ten['x']
        x1 = ggml_mul_mat(ctx, a, x, name=f'x1')
        x2 = ggml_add(ctx, x1, b, name=f'x2')
        return x2

    # create model graph
    model = GgmlCompute(params, tensors, test_model)

    # set weights
    a_dtype = np.float16 if qtype == T.F16 else np.float32
    a_np = np.random.randn(output_dim, input_dim).astype(a_dtype)
    b_np = np.random.randn(output_dim).astype(np.float32)
    model.set_input('a', a_np)
    model.set_input('b', b_np)

    # compute on input data
    x_np = np.random.randn(batch_size, input_dim).astype(np.float32)
    y_np = model(x=x_np)

    # get numpy results
    y0_np = (x_np @ a_np.T) + b_np[None,:]
    match = np.allclose(y_np, y0_np, atol=1e-5)

    # get rms and abs proportional errors
    rmse = np.sqrt(np.square(y_np-y0_np).mean()) / np.abs(y0_np).mean()
    abse = np.abs(y_np-y0_np).mean() / np.abs(y0_np).mean()
    print(match, rmse, abse)

    # return result
    return model

def test_torch(input_dim=256, output_dim=32, batch_size=16, qtype=T.F32, backend='cpu'):
    import torch
    from .ggml import ggml_mul_mat, ggml_add

    # model parameters
    params = dict(
        input_dim=input_dim, output_dim=output_dim, batch_size=batch_size
    )

    # tensor specifications
    tensors = dict(
        a = (qtype, (input_dim, output_dim)),
        b = (T.F32, (output_dim,)),
        x = (T.F32, (input_dim, batch_size)),
    )

    # define model function
    def test_model(ctx, par, ten):
        n, m = par['input_dim'], par['output_dim']
        a, b, x = ten['a'], ten['b'], ten['x']
        x1 = ggml_mul_mat(ctx, a, x, name=f'x1')
        x2 = ggml_add(ctx, x1, b, name=f'x2')
        return x2

    # create model graph
    model = GgmlCompute(params, tensors, test_model, backend=backend, framework='torch')

    # set weights
    a_dtype = torch.float16 if qtype == T.F16 else torch.float32
    a_pt = torch.randn(output_dim, input_dim).to(device=backend, dtype=a_dtype)
    b_pt = torch.randn(output_dim).to(device=backend, dtype=torch.float32)
    model.set_input('a', a_pt)
    model.set_input('b', b_pt)

    # compute on input data
    x_pt = torch.randn(batch_size, input_dim).to(device=backend, dtype=torch.float32)
    y_pt = model(x=x_pt)

    # get numpy results
    y0_pt = (x_pt @ a_pt.T.float()) + b_pt[None,:]
    match = torch.allclose(y_pt, y0_pt, atol=1e-5)

    # get rms and abs proportional errors
    rmse = (y_pt-y0_pt).square().mean().sqrt() / (y0_pt).abs().mean()
    abse = (y_pt-y0_pt).abs().mean() / (y0_pt).abs().mean()
    print(match, rmse.item(), abse.item())

    # return result
    return model
